
from .clientbase import ClientBase, run_client
from .common import SERVER_HOSTNAME, VIO_PORT

class VIOClient(ClientBase):
    '''Analog voltage input/output client.
    '''

    def __init__(self, host=None, port=None, running_index=0):

        if host is None:
            host = SERVER_HOSTNAME
        if port is None:
            port = int(VIO_PORT)
        super().__init__(host, port)


    def analog_input(self, duration, save=None, wait_trigger=False):
        '''Makes the server to record analog signal and save it.
        '''
        if save is None:
            save = 'None'
        self.send_command(
                f'analog_input;{duration}:{save}:{wait_trigger}')

    
    def set_settings(self, device, channels, fs):
        '''Configures the setttings in use.
        '''
        self.send_command(
                f'set_settings;{device}:{channels}:{fs}')
        
    
    def start_server(self):
        super().start_server('vio')
       

def main():
    run_client(VIOClient)

if __name__ == "__main__":
    main()



'''Commond code for all client programs.
'''
import socket
import time
import subprocess
import sys
import atexit

class ClientBase:
    '''Base class for all clients.

    The client offers a stable API for the main software (GonioImsoft)
    to use. The client internals and the server can be then more freely
    changed without causing havoc.

    Attributes
    ----------
    host : string
        The server IP (v4) address or a hostname
    port : int
        The server port number
    local_server : Popen obj or None
        None if no local server started by the client
    '''

    def __init__(self, host, port, running_index=0):
        self.host = host
        self.port = port


    def send_command(self, command, listen=False,
                     n_retry=60, retry_interval=1):
        '''Sends an arbitrary command to the server.

        Opens a connection to the server, send the command string and
        then optionally waits for the server's response.

        Arguments
        ---------
        command : string
            Format: "{command_name};{arg1}:{arg2}:..."
            Example 1: "ping;hello there"
            Example 2: "acquireSeries;0:0.1:0:5:label"
        listen : bool
            Wheter to wait and listen the server's response. If used
            incorrectly (listen=True but server says nothing, or
            listen=False and server says back), hangs either the client
            or the server.
        n_retry : int
            How many times to try to recontact the server if it cannot
            be reached (the server is off or the network is down etc.)
            Default is 60.
        retry_interval : int or float
            In seconds, how long to sleep between between the connection
            retries. Default is 1.
        '''
        if not isinstance(command, str):
            typ = type(command)
            raise TypeError(f'command must be a string, not {typ}')
        

        tries = 0
        host = self.host
        port = self.port

        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as soc:

            while True:
                try:
                    soc.connect((host, port))
                    break
                except ConnectionRefusedError:
                    tries += 1
                    if tries > n_retry:
                        raise ConnectionRefusedError(
                                f'Cannot connect to {host}:{port}')

                    print('Server connection unavailable, retrying...')
                    time.sleep(retry_interval)

            soc.sendall(command.encode())
            
            if listen:
                response = []
                while True:
                    data = soc.recv(1024)
                    if not data: break
                    response.append(data.decode())
                response = ''.join(response)
                if ':' in response:
                    response = response.split(':')
                return response


    def is_server_running(self):
        '''Returns True if the server responds to ping.
        '''
        try:
            self.send_command(
                    'ping;Hello there!', n_retry=0)
        except ConnectionRefusedError:
            return False
        return True

    def set_save_directory(self, directory):
        '''For any data saving that the server can do, set the directory.
        '''
        self.send_command(f'set_save_directory;{directory}')

    def start_server(self, name):
        '''Starts a local server if it is not running.

        Starts a new subprocess for the server.
        
        Arguments
        ---------
        name : string
            Name of the server. "camera" or "vio"
        '''
        if name not in ['camera', 'vio']:
            raise ValueError(
                    f'name has to be "camera" or "vio" not {name}')

        if self.is_server_running():
            print(f'Local server on port {self.port} already runs')
            print('-> Not starting another')
            return

        print(f'Starting a local server on port {self.port}')

        self.local_server = subprocess.Popen(
                [
                    sys.executable,
                    '-m', f'gonioimsoft.{name}_server',
                    '--port', str(self.port),
                    ],
                stdout=subprocess.DEVNULL)

        atexit.register(self.close_server)


    def close_server(self):
        '''Sens a closeure message for the server.

        If the server was started by this client (a local server), also
        waits the subprocess finish for 10 seconds unil terminates it.
        '''
        try:
            self.send_command('exit;parakalo', n_retry=0)
        except ConnectionRefusedError:
            pass
        
        if self.local_server is None:
            return

        atexit.unregister(self.close_server)

        if self.local_server is not None:
            try:
                self.local_server.wait(10)
            except subprocess.TimeoutExpired:
                self.local_server.terminate()
            self.local_server = None


def run_client(client):
    '''Runs the client from terminal without the main GonioImsoft program.
    '''
    import argparse

    parser = argparse.ArgumentParser(
            prog=f'GonioImsoft {client.__class__.__name__}',
            description=f'{client.__doc__}')

    parser.add_argument('-p', '--port')
    parser.add_argument('-a', '--address')

    args = parser.parse_args()

    client = client(args.address, args.port)

    print('Welcome to this interactive test program')
    print('Type in commands (or "help") and press enter')

    while True:
        cmd = input('#').split(' ')

        if not cmd:
            continue

        if cmd[0] == 'help':
            if len(cmd) == 1:
                help(client)
            else:
                method = getattr(client, cmd[1], None)
                if method is None:
                    print(f'No such command as "{cmd[1]}"')
                    continue
                
                print(method.__doc__)

        else:
            method = getattr(client, cmd[0], None)

            if method is None:
                print(f'No such command as "{cmd[0]}"')
                continue
            
            try:
                if len(cmd) == 1:
                    message = method()
                else:
                    message = method(*cmd[1:])
            except Exception as e:
                print('Given command casued an error in the client:')
                print(e)
                continue
            
            print(message)



'''Client code for the camera server/client division.
'''

import socket
import time
import os
import subprocess
import platform
import sys
import json
import atexit

from .directories import CODE_ROOTDIR, USERDATA_DIR
from .common import SERVER_HOSTNAME, CAMERA_PORT
from .clientbase import ClientBase, run_client

MAX_RETRIES = 100
RETRY_INTERVAL = 1

SAVEDIR = os.path.join(USERDATA_DIR, 'camera_states')


class CameraClient(ClientBase):
    '''Local part of the camera server/client division.

    CameraClient runs on the same PC as GonioImsoft and it connects to
    a CameraServer instance (over network sockets, so using IP addressess).
    It works as a middleman.
    
    No big data is transmitted over the connection, only commands (strings).
    It is the CameraServer's job to store the images, and display them on
    screen (livefeed) if needed.
    
    See also clientbase.py and camera_server.py for more information.

    Attributes
    ----------
    modified_settings : set
    '''


    def __init__(self, host=None, port=None, running_index=0):
        
        if host is None:
            host = SERVER_HOSTNAME
        if port is None:
            port = int(CAMERA_PORT) + int(running_index)
        super().__init__(host, port)

        self.modified_settings = set()
        self._roi = None


    def acquireSeries(self, exposure_time, image_interval, N_frames, label, subdir):
        '''
        Acquire a time series of images.
        For more see camera_server.py.

        Notice that it is important to give a new label every time
        or to change data savedir, otherwise images may be written over
        each other (or error raised).
        '''
        function = 'acquireSeries;'
        parameters = "{}:{}:{}:{}:{}".format(exposure_time, image_interval, N_frames, label, subdir)
        message = function+parameters
        
        self.send_command(message)


    def acquireSingle(self, save, subdir):
        self.send_command('acquireSingle;0.1:{}:{}'.format(str(save), subdir))
    
    def saveDescription(self, filename, string):
        self.send_command('saveDescription;'+filename+':'+string)

    def set_roi(self, roi):
        self._roi = roi
        self.send_command('set_roi;{}:{}:{}:{}'.format(*roi))

    def set_save_stack(self, boolean):
        self.send_command('set_save_stack;{}'.format(boolean))


    def get_cameras(self):
        '''Lists available cameras (their names) on the server.
        '''
        return self.send_command('get_cameras', listen=True)

    
    def get_camera(self):
        '''Returns a name describing the current camera device.
        '''
        return self.send_command('get_camera', listen=True)


    def set_camera(self, name):
        '''Sets what camera to use on the server.
        '''
        self.send_command(f'set_camera;{name}')


    def get_settings(self):
        '''Retrieves available settings of the camera device.
        '''
        return self.send_command('get_settings', listen=True)
    
    def get_setting_type(self, setting_name):
        '''Returns the type of the setting.
        One of the following: "string", "float" or "integer"
        '''
        return self.send_command(f'get_setting_type;{setting_name}',
                                listen=True)

    def get_setting(self, setting_name):
        '''Returns the current value of the setting as a string.
        '''
        return self.send_command(f'get_setting;{setting_name}',
                                listen=True)
    
    def set_setting(self, setting_name, value):
        '''Sets the specified setting to the specified value.
        '''
        self.send_command(f'set_setting;{setting_name}:{value}')
        self.modified_settings.add(setting_name)


    def save_state(self, label, modified_only=True):
        '''Acquires the current camera state and saves it
        
        modified_only : bool
            If True, save only those settings that have been edited
            by the user during this session.
        '''
        state = {}
        state['settings'] = {}
        
        # Save camera device settings
        for setting in self.get_settings():
            if modified_only and setting not in self.modified_settings:
                continue
            state['settings'][setting] = self.get_setting(setting)

        savedir = os.path.join(SAVEDIR, self.get_camera())
        os.makedirs(savedir, exist_ok=True)

        with open(os.path.join(savedir, f'{label}.json'), 'w') as fp:
            json.dump(state, fp)


    def load_state(self, label):
        '''Loads a previously saved camera state.
        '''

        savedir = os.path.join(SAVEDIR, self.get_camera())
        fn = os.path.join(savedir, f'{label}.json')
        
        if not os.path.exists(fn):
            raise FileNotFoundError(f'{fn} does not exist')

        with open(fn, 'r') as fp:
            state = json.load(fp)

        for setting, value in state['settings'].items():
            self.set_setting(setting, value)


    def list_states(self):
        '''Lists saved states available for the current camera.
        '''
        savedir = os.path.join(SAVEDIR, self.get_camera())
        if not os.path.isdir(savedir):
            return []
        return [fn.removesuffix('.json') for fn in os.listdir(savedir) if fn.endswith('.json')]


    def start_server(self):
        super().start_server('camera')

    def reboot(self):
        '''Performs a "reboot" for the camera and restores settings.

        Can be used as a "dirty fix" when the first image acqusition
        works fine but the subsequent ones crash for unkown reasons.
        '''
        self.set_camera(self.get_camera())
        self.load_state('previous')
        if self._roi:
            self.set_roi(self._roi)


def main():
    run_client(CameraClient)

if __name__ == "__main__":
    main()



'''Terminal user interface for GonioImsoft.

It uses a GonioImsoftCore instance to manage the experiments and
libtui.SimpleTUI to make the user interface.
'''

import os
import copy
import string
import time
import json
import inspect      # Inspect docs and source code

from gonioimsoft.version import __version__
from gonioimsoft.directories import (
        USERDATA_DIR,
        IS_USERDATA_INITIALIZED,
        initialize_userdata,
        )
from gonioimsoft.core import GonioImsoftCore, nidaqmx
from gonioimsoft.imaging_parameters import (
        DEFAULT_DYNAMIC_PARAMETERS,
        ParameterEditor,
        )
from .libtui import SimpleTUI


class Console:
    '''A command console for the terminal user interface.
    
    This console allows inputting commands with arguments.
    Needed, when simple keyboard shortcuts are not enough.

    Attributes
    ----------
    core : obj
        The GonioImsoftCore instance that this console operates on.
    '''
    def __init__(self, core):
        self.core = core


    def enter(self, user_input):
        '''Parses the user input and runs the command.
        '''
        command_name = user_input.split(' ')[0]
        args = user_input.split(' ')[1:]

        if hasattr(self, command_name):
            method = getattr(self, command_name)
            try:
                method(*args)
            except TypeError as e:
                print(e)
                self.help()
        else:
            print('Command {} does not exist'.format(command_name))
            self.help()
    

    def help(self, command_name=None):
        '''Prints help.

        Arguments
        ---------
        command_name : None or string
            The name of the command.
        '''
        if command_name is None:
            print('\n# List of commands:')
            
            for name, value in inspect.getmembers(self):
                
                if not inspect.ismethod(value):
                    continue

                helps = inspect.getdoc(value).split('\n')[0]

                print(f'  {name: >16}    {helps}')

            print('\nFor more instructions, try')
            print('  help [command_name]')
            print('  source [command name]')
        else:
            value = getattr(self, command_name, None)
            print()
            if value is not None:
                print(inspect.getdoc(value))
            print()

    def source(self, command_name):
        '''Prints the command's source code on screen (for help).
        '''
        val = getattr(self, command_name, None)
        if val is None:
            print(f'No command named {val}')
            return

        source = inspect.getsource(val)

        print(f'\nSource code of the "{command_name}" command (Python)\n')
        print(source)
        print('\nEnd of source.\n')

    def suffix(self, suffix):
        '''Set the suffix to add in the image folders' save name

        Arguments
        ---------
        suffix : string
            The appendix to the image folder name
        '''

        # Replaces spaces by underscores
        if ' ' in suffix:
            suffix = suffix.replace(' ', '_')
            print('Info: Replaced spaces in the suffix with underscores')
        
        # Replace illegal characters by x
        legal_suffix = ""
        for letter in suffix:
            if letter in string.ascii_letters+'_()-'+'0123456789.':
                legal_suffix += letter
            else:
                print('Replacing illegal character {} with x'.format(letter))
                legal_suffix += 'x'
        
        print('Setting suffix {}'.format(legal_suffix))
        self.core.set_subfolder_suffix(legal_suffix)


    def limitset(self, side, i_motor):
        '''Sets the current position as a limit for a motor.
        
        Arguments
        ---------
        side : string
            "upper" or "lower"
        i_motor : int
            Index of the motor.
        '''
        
        if side == 'upper':
            self.core.motors[i_motor].set_upper_limit()
        elif side == 'lower':
            self.core.motors[i_motor].set_lower_limit()
   

    def limitget(self, i_motor):
        '''
        Gets the current limits of a motor
        '''
        mlim = self.core.motors[i_motor].get_limits()
        print('  Motor {} limited at {} lower and {} upper'.format(i_motor, *mlim))


    def where(self, i_motor):
        '''Prints the coordinates of the motor i_motor.

        Arguments
        ---------
        i_motor : int
            Index of the motor
        '''
        # Getting motor's position
        mpos = self.core.motors[motor].get_position()
        print('  Motor {} at {}'.format(motor, mpos))


    def drive(self, i_motor, position):
        '''Drive i_motor to the given coordinates.

        Arguments
        ---------
        i_motor : int
            Index of the motor
        position : int or float
            The new position
        '''
        self.core.motors[i_motor].move_to(position)
        

    def macro(self, command, macro_name):
        '''Running and setting macros (automated imaging sequences).

        Arguments
        ---------
        command : string
            One of the following: run, list or stop
        macro_name : string
            The name of the macro.
        '''
        if command == 'run':
            self.core.run_macro(macro_name)
        elif command == 'list':

            print('Following macros are available')
            for line in self.core.list_macros():
                print(line)

        elif command == 'stop':
            for motor in self.core.motors:
                motor.stop()


    def set_roi(self, x,y,w,h, i_camera=None):
        '''Sets the camera crop or region of interest.

        Arguments
        ---------
        x, y, w, h : int
            Crop position and dimensions.
        i_camera : int or None
            The index of the camera to set the crop for.
            If not specified, uses the same crop for all the cameras.
        '''
        if i_camera is None:
            for camera in self.core.cameras:
                camera.set_roi((x,y,w,h))
        else:
            self.core.cameras[i_camera].set_roi( (x,y,w,h) )


    def eternal_repeat(self, isi):
        '''Repeats the imaging until the user hits enter.

        The save-suffix is appended with a running index.

        Arguments
        ---------
        isi : int
            In seconds, how long to wait between imagings
        '''

        isi = float(isi)
        print(isi)
        
        suffix = "eternal_repeat_isi{}s".format(isi)
        suffix = suffix + "_rep{}"
        i_repeat = 0
        
        while True:
            self.suffix(suffix.format(i_repeat))

            start_time = time.time()
            
            if self.core.image_series(inter_loop_callback=self.image_series_callback) == False:
                break
            i_repeat += 1

            sleep_time = isi - float(time.time() - start_time)
            if sleep_time > 0:
                time.sleep(sleep_time)


    def chain_presets(self, delay, *preset_names):
        '''Runs multiple presets all one after each other.

        The location (horizontal, vertical) should remain fixed.
        
        Arguments
        ---------
        delay : int or float
            In seconds, how long to wait between the presets.
        '''
        delay = float(delay)
        original_parameters = copy.copy(self.core.dynamic_parameters)

        
        print('Repeating presets {}'.format(preset_names))
        for preset_name in preset_names:
            print('Preset {}'.format(preset_name))
            
            self.core.load_preset(preset_name)
            
            if self.core.image_series(inter_loop_callback=self.image_series_callback) == False:
                break

            time.sleep(delay)

        print('Finished repeating presets')
        self.core.dynamic_parameters = original_parameters

            
    def set_rotation(self, horizontal, vertical):
        '''Sets the given rotation as the current one.
        '''
        ho = int(horizontal)
        ve = int(vertical)
        cho, cve = self.core.reader.latest_angle
        
        self.core.reader.offset = (cho-ho, cve-ve)


    def live(self):
        '''Toggles the cameras' livefeed (running/paused).
        '''
        if self.core.pause_livefeed == True:
            self.core.pause_livefeed = False
        else:
            self.core.pause_livefeed = True

    def violive(self, duration=None):
        '''Toggles the vios' livefeed (running/paused) or sets rec. dur.
        '''
        if duration is not None:
            duration = float(duration)
            self.core.vio_livefeed_dur = duration
        else:
            if self.core.vio_livefeed == True:
                self.core.vio_livefeed = False
            else:
                self.core.vio_livefeed = True


    def setoutput(self, device, channel, value):
        '''Sets an out-channel (eg. Dev1/ao1) to the given voltage value.
        '''
        try:
            self.core.set_led(f'{device}/{channel}', float(value))
        except Exception as e:
            print(e)


class GonioImsoftTUI:
    '''Terminal user interface for goniometric imaging.

    Attributes
    ----------
    core : obj
        The GonioImsoftCore instance
    console : object
    main_menu : list
        Main choices
    quit : bool
        If changes to True, quit.
    expfn : string
        Filename of the experiments.json file
    '''
    def __init__(self):
        
        self.libui = SimpleTUI()
        self.core = GonioImsoftCore()

        self.console = Console(self.core)
        self.console.image_series_callback = self.image_series_callback


        # Get experimenters list or if not present, use default
        self.expfn = os.path.join(USERDATA_DIR, 'experimenters.json')
        if os.path.exists(self.expfn):
            try:
                with open(self.expfn, 'r') as fp: self.experimenters = json.load(fp)
            except:
                self.experimenters = ['gonioims']
        else:
            self.experimenters = ['gonioims']
        


        self.experimenter = None    # name of the experimenter
        self.quit = False


    @property
    def main_menu(self):
        menu = [
                ['Normal imaging', self.loop_dynamic],
                ['Step-trigger imaging (takes an image on each goniometer step)', self.loop_static],
                ['Step-trigger (triggers only; use external camera software)', self.loop_trigger],
                ['\n', None],
                [f'Change savefolder (current: {self.experimenter})', self._run_experimenter_select],
                ['Quit', self.quit],
                ['\n', None],
                ['Add a local camera', self.add_local_camera],
                ['Add a remote camera', self.add_remote_camera],
                ['Edit camera settings', self.camera_settings_edit],
                ['Remove camera', self.remove_camera],
                ['\n', None],
                ['Add a local vio', self.add_local_vio],
                ['Add a remote vio', self.add_remote_vio],
                ['Remove vio', self.remove_vio],
                ]
        return menu


    def _add_camera(self, client):
        cameras = client.get_cameras()
        camera = self.libui.item_select(
                cameras, 'Select a camera')
        client.set_camera(camera)
        
        try:
            client.load_state('previous')
        except FileNotFoundError:
            self.libui.print('Could not find previous settings for this camera')
    
    def _add_vio(self, client):
        cancels = 'back'

        device = self.libui.input('Device', cancels)
        channels = self.libui.input('Channels (comma separated)', cancels)
        fs = self.libui.input('Sampling frequency (Hz)', cancels)

        client.set_settings(device, channels, fs)

    def add_local_camera(self):
        '''Add a camera from a local camera server.
        '''
        client = self.core.add_camera_client(None, None)
        self._add_camera(client)


    def add_local_vio(self):
        '''Start a local vio server and a client.
        '''
        client = self.core.add_vio_client(None, None)
        self._add_vio(client)


    def _add_remote_client(self, name):
        if name == 'camera':
            addfunc = self.core.add_camera_client
        elif name == 'vio':
            addfunc = self.core.add_vio_client
        else:
            raise ValueError

        cancels = 'back'
        self.libui.print(f'# Type in {cancels} to cancel')

        host = self.libui.input('IP address or hostname', cancels)
        if host is None:
            return
        port = self.libui.input('Port (leave blank for default): ', cancels)
        if port is None:
            return

        if port == '':
            port = None
        else:
            port = int(port)

        client = addfunc(host, port)
        
        while not client.is_server_running():
            print('Waiting the server to come up...')
            time.sleep(1)

        if not client.is_server_running():
            self.libui.print('Cannot connect to the server')
        else:
            if name == 'camera':
                self._add_camera(client)
            else:
                self._add_vio(client)

    
    def add_remote_camera(self):
        self._add_remote_client('camera')
    
    def add_remote_vio(self):
        self._add_remote_client('vio')



    def remove_camera(self):
        names = [cam.get_camera() for cam in self.core.cameras]
        selection = self.libui.item_select(
                names+['..back'], 'Select camera to remove')

        if selection != '..back':
            index = names.index(selection)
            self.core.remove_camera_client(index)

    def remove_vio(self):
        names = [f'vio_{i}' for i, vio in enumerate(self.core.vios)]
        selection = self.libui.item_select(
                names+['..back'], 'Select camera to remove')

        if selection != '..back':
            index = names.index(selection)
            self.core.remove_vio_client(index)

   

    @property
    def menutext(self):
        cam = ''

        # Check camera server status
        for i_camera, camera in enumerate(self.core.cameras):
            if camera.is_server_running():
                cam_name = camera.get_camera()
                if cam_name:
                    cs = f'{cam_name}\n'
                else:
                    cs = 'No camera selected\n'
            else:
                cs = 'Offline'

            cam += f'Cam{i_camera} {cs}'

        if not self.core.cameras:
            cam = 'No cameras'

        # Check serial (Arduino) status
        ser = self.core.reader.serial
        if ser is None:
            ar = 'Serial UNAVAIBLE'
        else:
            if ser.is_open:
                ar = 'Serial OPEN ({} @{} Bd)'.format(
                        ser.port, ser.baudrate)
            else:
                ar = 'Serial CLOSED'

        # Check DAQ
        if nidaqmx is None:
            daq = 'UNAVAILABLE'
        else:
            daq = 'AVAILABLE'

        status = "\n {} | {} | nidaqmx {}".format(cam, ar, daq)
        
        menutext = "GonioImsoft - Version {}".format(__version__)
        menutext += "\n" + max(len(menutext), len(status)) * "-"
        menutext += status
        return menutext + "\n"


    def loop_trigger(self):
        '''
        Simply NI trigger when change in rotatory encoders, leaving camera control
        to an external software (the original loop static).
        
        Space to toggle triggering.
        '''
        self.loop_dynamic(static=True, camera=False)
                

    def loop_static(self):
        '''
        Running the static imaging protocol.
        '''
        self.loop_dynamic(static=True)
        

    def image_series_callback(self, label, i_repeat):
        '''
        Callback passed to image_series
        '''
        if label:
            print(label)
        
        key = self.libui.read_key()

        if key == '\r':
            # If Enter presed return False, stopping the imaging
            print('User pressed enter, stopping imaging')
            return False
        else:
            return True


    def loop_dynamic(self, static=False, camera=True):
        '''
        Running the dynamic imaging protocol.

        static : bool
            If False, run normal imaging where pressing space runs the imaging protocol.
            If True, run imaging when change in rotary encoders (space-key toggled)
        camera : bool
            If True, control camera.
            If False, assume that external program is controlling the camera, and send trigger
        '''
        trigger = False
        
        self.core.set_savedir(os.path.join('imaging_data_'+self.experimenter), camera=camera)

        cancels = 'back'
        self.libui.print(f'# Enter specimen metadata (enter {cancels} to cancel)\n')

        name = self.libui.input(
                'Name ({})'.format(self.core.preparation['name']), cancels)
        if name is None: return

        sex = self.libui.input(
                'Sex ({})'.format(self.core.preparation['sex']), cancels)
        if sex is None: return

        age = self.libui.input(
                'Age ({})'.format(self.core.preparation['age']), cancels)
        if age is None: return

        if self.core.initialize(name, sex, age, camera=camera) is None:
            return

        upper_lines = ['-','Dynamic imaging', '-', 'Help F1', 'Space ']

        self.libui.clear_screen()


        help_string = "# This part of the program works by keyboard shortcuts\n"
        if static:
            help_string += '#   space      Toggle trigger-by-rotation on/off\n'
        else:
            help_string += "#   space      Run imaging\n"
        help_string += (
                '#   enter      Return back to the main menu\n'
                "#   0          Set current rotation as (0,0)\n"
                "#   s          Take a snap image\n"
                '#   e          Edit imaging parameters again\n'
                '#   h          Print this help again\n'
                "#   ` (tilde)  Open command console (type in help for help)\n"
                "# \n"
                "# Rotation stage changes will be printed here.\n"
                )
        if not self.core.cameras:
            help_string += (
                    '# You have not added any cameras. Space now triggers only.\n'
                    '# Return to the main menu to add cameras.\n'
                    )
        else:
            help_string += (
                    '# Separate windows for the added cameras should open\n'
                    )


        self.libui.print(help_string)

        while True:
            
            lines = upper_lines

            key = self.libui.read_key()

            if static:
                if trigger and self.core.trigger_rotation:
                    if camera:
                        self.core.image_series(inter_loop_callback=self.image_series_callback)
                    else:
                        self.core.send_trigger()
                if key == ' ':
                    trigger = not trigger
                    print('Rotation triggering now set to {}'.format(trigger))
            else:
                if key == ' ':
                    if camera:
                        self.core.image_series(inter_loop_callback=self.image_series_callback)
                    else:
                        self.core.send_trigger()
            
            if key == 112:
                lines.append('')
            elif key == '0':
                self.core.set_zero()
            elif key == 's':
                if camera:
                    self.core.take_snap(save=True)
            elif key == '\r':
                # If user hits enter we'll exit
                break
            elif key == 'e':
                if self.core.initialize(name, sex, age, camera=camera) is None:
                    continue
            elif key == 'h':
                self.libui.print(help_string)
            elif self.core.motors:
                if key == '[':
                    self.core.motors[0].move_raw(-1)
                elif key == ']':
                    self.core.motors[0].move_raw(1)
                
                elif key == 'o':
                    self.core.motors[1].move_raw(-1)
                elif key == 'p':
                    self.core.motors[1].move_raw(1)

                elif key == 'l':
                    self.core.motors[2].move_raw(-1)
                elif key == ';':
                    self.core.motors[2].move_raw(1)
            elif key == '`':
                user_input = self.libui.input("Type command >> ", '')
                if user_input is None:
                    continue
                self.console.enter(user_input)

            elif key == '' and not (static and self.core.trigger_rotation):
                if camera:
                    # When there's no input just update the live feed
                    self.core.take_snap(save=False)
            
            
            #self._clearScreen()
            #self._print_lines(lines)

            self.core.tick()

        self.core.finalize()


    
    def _run_firstrun(self):
        message = (
                '\nHello and welcome! This is your first run.\n'
                '\n'
                'GonioImsoft needs a location '
                'to save user files\n- list of savefolders\n'
                '- camera states\n'
                '- macros\n'
                '- presets\n'
                '\n'
                f'The location is {os.path.abspath(USERDATA_DIR)}\n'
                'No imaging data or big files will be saved here.\n'
                f'\nCreate {USERDATA_DIR}? (yes recommended)\n'
                )
        if self.libui.bool_select(message):
            initialize_userdata()
            print('Success!')
            time.sleep(2)
        else:
            print('Warning! Cannot save any changes')
            time.sleep(2)


    def _run_experimenter_select(self):

        if self.experimenter is not None:
            self.libui.print(f'Current savefolder: {self.experimenter}')

        extra_options = [' (Add new)', ' (Remove old)', ' (Save current list)']

        self.libui.print('Select savefolder\n--------------------')
        while True:
            # Select operation
            selection = self.libui.item_select(self.experimenters+extra_options) 

            self.libui.clear_screen()
            
            # add new
            if selection == extra_options[0]:
                cancels = 'back'
                self.libui.print(f'# Adding new user (enter {cancels} to cancel)')
                name = self.libui.input('Name >>', cancels)
                if name is None:
                    continue
                self.experimenters.append(name)

            # remove old
            elif selection == extra_options[1]:
                self.libui.print('Select who to remove (data remains)')
                
                to_delete_name = self.libui.item_select(
                        self.experimenters+['..back (no deletion)'])

                if to_delete_name in self.experimenters:
                    self.experimenters.pop(self.experimenters.index(to_delete_name))

            # save current
            elif selection == extra_options[2]:
                if os.path.isdir(USERDATA_DIR):
                    with open(self.expfn, 'w') as fp: json.dump(self.experimenters, fp)
                    print('Saved!')
                else:
                    print(f'Saving failed (no {USERDATA_DIR})')
                time.sleep(2)
            else:
                # Got a name
                break

            self.libui.clear_screen()

        self.experimenter = selection
 

    def run(self):
        '''
        Run TUI until user quitting.
        '''
        
        self.libui.header = self.menutext
        self.libui.clear_screen()
 
        # Check if userdata directory settings exists
        # If not, ask to create it
        if not IS_USERDATA_INITIALIZED:
            self._run_firstrun()
            self.libui.clear_screen()
        else:
            # Already initialized, just check all subfolders present
            # in newer versions are also there
            initialize_userdata()
        
        self._run_experimenter_select()
        self.libui.clear_screen()

        self.quit = False
        while not self.quit:
            self.libui.clear_screen()
            
            menuitems = [x[0] for x in self.main_menu]
            
            # Blocking call here
            selection = self.libui.item_select(menuitems)

            self.libui.clear_screen()
            self.main_menu[menuitems.index(selection)][1]()

            # Update status menu and clear screen
            self.libui.header = self.menutext

        self.core.exit()

    
    def camera_settings_edit(self):
        '''View to select a camera and edit it's settings
        '''
        
        while True:

            camera = self.libui.item_select(
                    self.core.cameras+['..back'],
                    "Select the camera to edit")
            
            if camera == '..back':
                break
            
            while True:
                setting_name = self.libui.item_select(
                        camera.get_settings()+['..back'],
                        "Select the setting to edit")

                if setting_name == '..back':
                    break
                
                value = camera.get_setting(setting_name)
                value_type = camera.get_setting_type(setting_name)

                self.libui.print(f'{setting_name} ({value_type})')
                self.libui.print(f'Current value: {value}')
                new_value = self.libui.input('New value: ')

                camera.set_setting(setting_name, new_value)
                
                self.libui.clear_screen()
                
                camera.save_state('previous')

    def quit(self):
        self.quit = True



def main():
    imsoft = GonioImsoftTUI()
    imsoft.run()

if __name__ == "__main__":
    main()



'''GonioImsoft main components.
'''

import os
import sys
import time
import datetime
import copy

import numpy as np

try:
    import nidaqmx
except ModuleNotFoundError:
    nidaqmx = None

from gonioimsoft.anglepairs import saveAnglePairs, loadAnglePairs, toDegrees
from gonioimsoft.arduino_serial import ArduinoReader
from gonioimsoft.camera_client import CameraClient
from gonioimsoft.vio_client import VIOClient
from gonioimsoft.motors import Motor
from gonioimsoft.imaging_parameters import (
        DEFAULT_DYNAMIC_PARAMETERS,
        load_parameters,
        getModifiedParameters)
from gonioimsoft.stimulus import StimulusBuilder
import gonioimsoft.macro as macro

ENABLE_MOTORS = False

class GonioImsoftCore:
    '''Main interface to control GonioImsoft recordings.

    Attributes
    ----------
    reader : obj
        Reading rotary encoder values from the Arduino Board.
    cameras : list
        Camera client objects that each talk to a specific server
        over IPv4 sockets. The server's can be local (on the same
        machine) or remote.
    vios : list
        Analog voltage input/ouput clients. Similar to the cameras,
        they talk to a vio server.
    '''


    def __init__(self, dynamic_parameters=DEFAULT_DYNAMIC_PARAMETERS):
        '''
        Sets up ArduinoReader, CameraClient/Server.
        '''

        # Angle pairs reader (rotary encoder values)
        self.reader = ArduinoReader()
        
        self.cameras = []
        self.vios = []
        
        # Details about preparation (name, sex, age) are saved in this
        self.preparation = {'name': 'test', 'sex': '', 'age': ''}

        self.dynamic_parameters = dynamic_parameters

        self.previous_angle = None

        # Suffix to be appended in the subfolders, see set_subfolder_suffix method
        self.suffix = ''

        # Set up motors followingly
        # 0)    Horizontal + sensor
        # 1)    Vertical + sensor
        # 2)    Microscope focus (no sensor)
        self.motors = []
        if ENABLE_MOTORS:
            for i_motor, i_sensor in zip([0,1,2],[0,1,None]):
                self.motors.append(Motor(self.reader, i_motor, i_sensor))
        

        # Macro imaging: Automatically move motors and take images
        # self macro is a list of anglepairs where to move the horizontal/vertical,
        # take image "commands" (string 'image') and other functions.
        # For more details see the tick method of this class.
        self.macro = None
        self.i_macro = 0
        self.waittime = 0

        self.trigger_rotation = False

        self.trigger_signal = np.array([3,3,3,0])
        self.triggered_anglepairs = None

        self.data_savedir = None

        self.local_camera_servers_running_index = 0
        self.local_vio_servers_running_index = 0

        self.pause_livefeed = False
        self.vio_livefeed = False
        self.vio_livefeed_dur = 0.1

        self._last_vio = time.time()

    
    def _add_client(self, name, host, port):
        '''Adds a camera client to the given host and port.

        If host is None uses the localhost and starts a local
        server if no local server running at that port

        Arguments
        ---------
        name : string
            The name of the client. "camera" or "vio"
        '''
        if name == 'camera':
            Client = CameraClient
            register = self.cameras
            self.local_camera_servers_running_index += 1
            index = self.local_camera_servers_running_index
        elif name == 'vio':
            Client = VIOClient
            register = self.vios
            self.local_vio_servers_running_index += 1
            index = self.local_vio_servers_running_index
        else:
            raise ValueError

        client = Client(host, port, running_index=index-1)

        if host is None and not client.is_server_running():
            client.start_server()
        register.append(client)
        
        return client


    def add_camera_client(self, host, port):
        return self._add_client('camera', host, port)
        
    def add_vio_client(self, host, port):
        return self._add_client('vio', host, port)


    def _remove_client(self, name, i_client):
        if name == 'camera':
            register = self.cameras
        elif name == 'vio':
            reigster = self.vios
        # Popping is enough and the client should be garbage collected
        # by Python (the sockets are not kept alive so nothing is
        # left open etc. by the client)
        client = self.cameras.pop(i_client)
        
        # If the client started a local server, close the server
        if client.local_server is not None:
            client.close_server()

        return client

    def remove_camera_client(self, i_client):
        '''Removes the camera client and closes its server if local
        '''
        return self._remove_client('camera', i_client)
   

    def remove_vio_client(self, i_client):
        '''Removes the vio client and closes its server if local
        '''
        return self._remove_client('vio', i_client)



    def analog_output(self, channels, stimuli, fs, wait_trigger, camera=True):
        '''

        channels    List of channel names
        stimuli     List of 1D numpy arrays
        fs          Sampling frequency of the stimuli
        camera : bool
            If True, send the camera server a "ready" command
        '''
        
        if nidaqmx is None:
            print('    pretending analog_output on channels {}'.format(channels))
            return None


        with nidaqmx.Task() as task:
            for i_channel, channel in enumerate(channels):
                if type(channel) == type('string'):
                    task.ao_channels.add_ao_voltage_chan(channel)
                else:
                    for subchan in channel:
                        task.ao_channels.add_ao_voltage_chan(subchan)
                        stimuli.insert(i_channel, stimuli[i_channel])
                    stimuli.pop(i_channel)
                    
            task.timing.cfg_samp_clk_timing(float(fs), samps_per_chan=len(stimuli[0]))

            
            if len(stimuli) > 1:
                stimulus = stimuli[0]
                for s in stimuli[1:]:
                    stimulus = np.vstack((stimulus, s))
            else:
                stimulus = stimuli[0]

            task.write(stimulus)

            if wait_trigger:
                task.triggers.start_trigger.cfg_dig_edge_start_trig("/Dev1/PFI0", trigger_edge=nidaqmx.constants.Edge.RISING)

            
            task.start()

            if camera:
                for camera in self.cameras:
                    camera.send_command('ready')
                
            task.wait_until_done(timeout=(len(stimuli[0])/fs)*1.5+20)


    def send_trigger(self):
        '''
        Sending trigger.
        '''

        chan = self.dynamic_parameters.get('trigger_out_channel', None)
        
        if nidaqmx is None:
            print(f'    pretending to send trigger on chan {chan}')
            return None

        if chan:
            with nidaqmx.Task() as task:
                task.ao_channels.add_ao_voltage_chan(chan)
                task.timing.cfg_samp_clk_timing(1000., samps_per_chan=4)
                task.write(self.trigger_signal)
                task.start()
                task.wait_until_done(timeout=1.)
        else:
            print('trigger_out_channel not specified, no trigger out')

        self.triggered_anglepairs.append(self.reader.get_latest())


    def set_led(self, device, value, wait_trigger=False, exclude=None):
        '''
        Set an output channel to a specific voltage value.

        INPUT ARGUMENTS     DESCRIPTION
        device              A string (single device) or a list of strings (many devices at once)
        '''
        if nidaqmx is None:
            print(f'    pretending to set {device} on value {value}')
            return None


        excluded = 0

        with nidaqmx.Task() as task:
            
            if type(device) == type('string'):
                # If there's only a single device
                task.ao_channels.add_ao_voltage_chan(device)
            else:
                # If device is actually a list of devices
                for dev in device:
                    if exclude and dev in exclude:
                        excluded += 1
                        continue
                    task.ao_channels.add_ao_voltage_chan(dev)
                value = [value for i in range(len(device)-excluded)]
            
            if wait_trigger:
                task.timing.cfg_samp_clk_timing(10000)
                task.triggers.start_trigger.cfg_dig_edge_start_trig("/Dev1/PFI0", trigger_edge=nidaqmx.constants.Edge.FALLING)
            task.write(value)
            


    def wait_for_trigger(self):
        '''
        Doesn't return until trigger signal is received.
        '''
        if nidaqmx is None:
            print(f'    waiting for trigger')
            return None

        with nidaqmx.Task() as task:
            device = nidaqmx.system.device.Device('Dev1')
            
            task.ai_channels.add_ai_voltage_chan('Dev1/ai0' )
            
            task.timing.cfg_samp_clk_timing(10000)
            
            task.triggers.start_trigger.cfg_dig_edge_start_trig("/Dev1/PFI0", trigger_edge=nidaqmx.constants.Edge.RISING)
            task.read(number_of_samples_per_channel=1)
        


    #
    # IMAGING METHODS
    #

    def take_snap(self, save=False):
        '''
        Takes a snap image.
        
        save        Whether to save the image directly or not.
        '''
        if save:
            self.set_led(self.dynamic_parameters['ir_channel'], self.dynamic_parameters['ir_imaging'])
            time.sleep(0.3)
            for camera in self.cameras:
                camera.acquireSingle(True, os.path.join(self.preparation['name'], 'snaps'))
            self.set_led(self.dynamic_parameters['ir_channel'], self.dynamic_parameters['ir_livefeed'])
            time.sleep(0.2)

            print('A snap image taken')
        else:
            if self.pause_livefeed:
                return


                
            for camera in self.cameras:
                camera.acquireSingle(False, '')
            time.sleep(0.1)





    
    def image_trigger_softhard(self):
        '''Software triggering of the cameras.

        How this works?
        CameraClient (self.camera) send message to CameraServer to start image acquisition. Starting
        image acquistion takes some time, so to synchronize, we wait a trigger signal from the camera
        before turning the LED on. What is done is essentially half software half hardware triggering.
        Not ideal but seems to work.
        '''
        
        self.set_led(dynamic_parameters['ir_channel'], dynamic_parameters['ir_imaging'])
        time.sleep(0.5)
        
        # Subfolder suffix so if experimenter takes many images from the same position in different conditions
        self.camera.acquireSeries(frame_length, 0, N_frames, label, os.path.join(self.preparation['name'], 'pos{}_{}'.format(imaging_angle, dynamic_parameters['suffix']+self.suffix)), 'send')
        
        self.wait_for_trigger()
        time.sleep(dynamic_parameters['pre_stim'])
        
        self.set_led(dynamic_parameters['flash_channel'], dynamic_parameters['flash_on'][i])
        time.sleep(dynamic_parameters['stim'])

        self.set_led(dynamic_parameters['flash_channel'], dynamic_parameters['flash_off'])
        
        time.sleep(dynamic_parameters['post_stim'])
        
        self.set_led(dynamic_parameters['ir_channel'], dynamic_parameters['ir_waiting'])
        

    def image_trigger_hard_cameraslave(self):
        '''
        Where camera is triggered by square wave.
        Since the camera cannot be run this way at 100Hz at full frame,
        image_series3 is used instead.
        '''

        fs = 1000

        stimulus, illumination, camera = get_pulse_stimulus(dynamic_parameters['stim'],
                dynamic_parameters['pre_stim'], dynamic_parameters['post_stim'],
                dynamic_parameters['frame_length'], dynamic_parameters['flash_on'][i],
                dynamic_parameters['ir_imaging'], fs,
                stimulus_finalval=dynamic_parameters['flash_off'],
                illumination_finalval=dynamic_parameters['ir_waiting'])
        

        self.camera.acquireSeries(frame_length, 0, N_frames, label,
                os.path.join(self.preparation['name'], 'pos{}{}'.format(imaging_angle,
                    dynamic_parameters['suffix']+self.suffix)), 'receive')
        
        time.sleep(5)

        
        self.analog_output([dynamic_parameters['flash_channel'],
                dynamic_parameters['ir_channel'],
                dynamic_parameters['trigger_channel']],
                [stimulus, illumination, camera], fs)

      

    def image_series(self, trigger='hard_cameramaster', inter_loop_callback=None):
        '''
        Contains common steps for all image_series methods.
        
        triggering_type     "softhard", "hard_cameraslave", "hard_cameramaster"

        Returns True if finished properly and False if user cancelled.
        '''
        
        
        
        exit_imaging = False

        print('Starting dynamic imaging using {} triggering'.format(trigger))
        if trigger == 'softhard':
            imaging_function = self.image_trigger_softhard
        elif trigger == 'hard_cameraslave':
            imaging_function = self.image_trigger_hard_cameraslave
        elif trigger == 'hard_cameramaster':
            imaging_function = self.image_trigger_hard_cameramaster

        # Wait ISI period here, if it has not passed since the last series imaging
        # Otherwise, continue.
        try:
            if time.time() < self.isi_slept_time: 
                print('Waiting ISI to be fullfilled from the last run...')
                time.sleep(self.isi_slept_time - time.time())
                print('READY')
        except AttributeError:
            pass
        
        dynamic_parameters = copy.deepcopy(self.dynamic_parameters)

        # Check that certain variables are actually lists (used for intensity series etc.)
        # Check also for correct length if a list
        for param in ['isi', 'flash_on']:
            if type(dynamic_parameters[param]) != type([]):
                dynamic_parameters[param] = [dynamic_parameters[param]] * dynamic_parameters['repeats']
            
            elif len(dynamic_parameters[param]) != int(dynamic_parameters['repeats']):
                print('Warning! Dynamic parameter {} length is {} but repeats is set to {}'.format(param,
                    len(dynamic_parameters[param]), dynamic_parameters['repeats'] ))
                dynamic_parameters[param] = [dynamic_parameters[param][0]] * dynamic_parameters['repeats'] 

        # Set stack save option
        for camera in self.cameras:
            camera.set_save_stack(dynamic_parameters.get('save_stack', False))
        

        # Get the current rotation stage angles and use this through the repeating
        # (even if it would change during imaging)
        imaging_angle = self.reader.get_latest()

        spaceless_angle = str(imaging_angle).replace(' ', '')
        
        # Prepare some variables that stay constant over imaging
        image_directory = os.path.join(self.preparation['name'], 'pos{}{}'.format(spaceless_angle, dynamic_parameters['suffix']+self.suffix))
        N_frames = int((dynamic_parameters['pre_stim']+dynamic_parameters['stim']+dynamic_parameters['post_stim'])/dynamic_parameters['frame_length'])
       

        for i in range(dynamic_parameters['repeats']):

            label = 'im_pos{}_rep{}'.format(spaceless_angle, i)
            
            # INTER_LOOP_CALLBACK for showing info to the user and for exiting
            if callable(inter_loop_callback) and inter_loop_callback(label, i) == False:
                exit_imaging = True  
            if exit_imaging:
                break
            
            fs = 1000
            builder = StimulusBuilder(dynamic_parameters['stim'],
                dynamic_parameters['pre_stim'], dynamic_parameters['post_stim'],
                dynamic_parameters['frame_length'], dynamic_parameters['flash_on'][i],
                dynamic_parameters['ir_imaging'], fs,
                stimulus_finalval=dynamic_parameters['flash_off'],
                illumination_finalval=dynamic_parameters['ir_waiting'],
                wtype=dynamic_parameters['flash_type'])

            if dynamic_parameters.get('biosyst_stimulus', ''):
                bsstim, fs = builder.overload_biosyst_stimulus(dynamic_parameters['biosyst_stimulus'], dynamic_parameters['biosyst_channel'])
                N_frames = int(round((len(bsstim)/fs) / dynamic_parameters['frame_length']))

            if i==0 and dynamic_parameters['avgint_adaptation']:
                self.set_led(dynamic_parameters['flash_channel'], np.mean(builder.get_stimulus_pulse()), exclude='Dev1/ao4')
                time.sleep(dynamic_parameters['avgint_adaptation'])
            
            imaging_function(dynamic_parameters, builder, label, N_frames, image_directory, set_led=bool(dynamic_parameters['isi'][i]))

            if i==0 and dynamic_parameters['avgint_adaptation']:
                self.set_led(dynamic_parameters['flash_channel'], np.mean(builder.get_stimulus_pulse()), exclude='Dev1/ao4')

            # Dirtyfix
            if dynamic_parameters['reboot_cameras']:
                print("Rebooting cameras")
                for i_camera, camera in enumerate(self.cameras):
                    print(f'  cam_{i_camera}...')
                    camera.reboot()

            # Wait the total imaging period; If ISI is short and imaging period is long, we would
            # start the second imaging even before the camera is ready
            # Better would be wait everything clear signal from the camera.
            #total_imaging_time = dynamic_parameters['pre_stim'] + dynamic_parameters['stim'] + dynamic_parameters['post_stim']
            #print("Total imaging time " + str(total_imaging_time))
            # Do the actual waiting together with ISI, see just below
            
            # WAITING ISI PERIOD
            if i+1 == dynamic_parameters['repeats']:
                self.isi_slept_time = time.time() + dynamic_parameters['isi'][i]
            else:
                wakeup_time = time.time() + dynamic_parameters['isi'][i] #+ total_imaging_time
                
                while wakeup_time > time.time():
                    if callable(inter_loop_callback) and inter_loop_callback(None, i) == False:
                        exit_imaging = True
                        break
                    time.sleep(0.01)

        self.set_led(dynamic_parameters['flash_channel'], dynamic_parameters['flash_off'])
        self.set_led(dynamic_parameters['ir_channel'], dynamic_parameters['ir_livefeed'])
        print('DONE!')

        if exit_imaging:
            return False
        else:
            return True


    def image_trigger_hard_cameramaster(self, dynamic_parameters, builder, label, N_frames, image_directory, set_led=True):
        '''
        When starting the imaging, the camera sends a trigger pulse to NI board, leading to onset
        of the stimulus (hardware triggering by the camera).
        
        Illumination IR light is hardware triggered together with the stimulus.

        set_led   Set IR to ir_waiting in between (if long enough ISI)
        '''
        if set_led:
            self.set_led(dynamic_parameters['ir_channel'], dynamic_parameters['ir_imaging'])
            time.sleep(0.5)
        
        fs = builder.fs
          
        stimulus = builder.get_stimulus_pulse()
        if isinstance(stimulus, list):
            NN = len(stimulus[0])
        else:
            NN = stimulus.shape
        irwave = dynamic_parameters['ir_imaging'] * np.ones(NN)

        #irwave = dynamic_parameters['ir_imaging'] * np.ones(stimulus.shape)
        if set_led:
            irwave[-1] = dynamic_parameters['ir_waiting']

        stimuli = []
        channels = []

        if isinstance(stimulus, list):
            # Many stimulus channels
            print('Many stimulus channels')
            stimuli = [*stimulus, irwave]
            channels = [*dynamic_parameters['flash_channel'], dynamic_parameters['ir_channel']]
        else:
            # One stimulus channel
            stimuli = [stimulus, irwave]
            channels = [dynamic_parameters['flash_channel'], dynamic_parameters['ir_channel']]



        # Arm analog input recording if any vio clients
        for i_vio, vio in enumerate(self.vios):
            vio.set_save_directory(os.path.join(self.data_savedir, image_directory))
            duration = N_frames * dynamic_parameters['frame_length']
            vio_label = f'vi{i_vio}{label[2:]}'
            vio.analog_input(duration, save=vio_label, wait_trigger=True)


        if len(self.cameras) == 1:
            self.cameras[0].acquireSeries(dynamic_parameters['frame_length'], 0, N_frames, label, image_directory)
        elif len(self.cameras) > 1:
            # With many cameras, add camN suffix to the label
            for i_camera, camera in enumerate(self.cameras):
                camera.acquireSeries(dynamic_parameters['frame_length'], 0, N_frames, f'{label}_cam{i_camera}', image_directory)


        # If no cameras, we should not wait for trigger to come from them.
        # Create own trigger on the trigger channel
        if not self.cameras:
            wait_trigger = False
            trigwave = np.zeros(len(stimuli[0]))
            for i in range(min(100, len(trigwave))):
                trigwave[i] = 5
            
            stimuli = [*stimuli, trigwave]
            channels = [*channels, dynamic_parameters['trigger_out_channel']]
        else:
            wait_trigger = True
        
        self.analog_output(channels, stimuli, fs, wait_trigger=wait_trigger)

        


    def set_savedir(self, savedir, camera=True):
        '''
        Set the directory where the taken images are saved.

        camera : bool
            If False, do not attempt to update save folder to the camera server
        '''
        if camera:
            for device in self.cameras+self.vios:
                device.set_save_directory(savedir)
        self.data_savedir = savedir


    def set_subfolder_suffix(self, suffix):
        '''
        Set any suffix to a data folder containing the images.
        For example
            pos(-14,0) would become pos(-14,0)_highmag if suffix == "highmag"
        '''
        if suffix:
            self.suffix = '_'+suffix
        else:
            self.suffix = ''

    def _update_descriptions_file(self):
        # Saving description file
        desc_string = "name {}\nsex {}\nage {}".format(self.preparation['name'], self.preparation['sex'], self.preparation['age'])
        desc_string += "\n\n#DYNAMIC PROTOCOL PARAMETERS\n"
        for name, value in self.dynamic_parameters.items():
            desc_string += '{} {}\n'.format(name, value)
        for camera in self.cameras:
            camera.saveDescription(self.preparation['name'], desc_string)
        

    def initialize(self, name, sex, age, camera=False):
        '''Call this to initialize the experiments.

        Returns True if worked or None if user cancelled.


        name, sex age       Can be '' (empty string)
        '''
        # Preparations, ask droso name
        if name != '':
            self.preparation['name'] = name
        if sex != '':
            self.preparation['sex'] = sex
        if age != '':
            self.preparation['age'] = age

        params = getModifiedParameters()
        if params is None:
            return None 

        self.dynamic_parameters = params

        if camera:
            self._update_descriptions_file()
        else:
            self.triggered_anglepairs = []
        
        self.set_led(self.dynamic_parameters['ir_channel'], self.dynamic_parameters['ir_livefeed'])
        self.set_led(self.dynamic_parameters['flash_channel'], self.dynamic_parameters['flash_off'])
       
        roi = self.dynamic_parameters['ROI']
        if roi is not None:
            for camera in self.cameras:
                camera.set_roi(roi)

        return True

    def load_preset(self, preset_name):
        fn = os.path.join('presets', preset_name)
        self.dynamic_parameters = load_parameters(fn)
        self._update_descriptions_file()



    def tick(self, horizontal_trigger=True, vertical_trigger=False):
        '''Updates the current angle and performs all "houskeeping" duties also.
        
        Meant to be called repeatedly inside the UI loop.
        '''

        change = False

        if self.vio_livefeed:
            if time.time()-self._last_vio > max(self.vio_livefeed_dur+0.1, 0.1):
                for vio in self.vios:
                    vio.analog_input(self.vio_livefeed_dur)
                self._last_vio = time.time()
        
        while True:
            
            # Update current angle and echo it to the console
            current_angle = [list(self.reader.read_angles())]
            toDegrees(current_angle)

            if self.previous_angle != current_angle:
                horchanged = self.previous_angle and self.previous_angle[0][0] != current_angle[0][0]
                verchanged = self.previous_angle and self.previous_angle[0][1] != current_angle[0][1]
                if (horizontal_trigger and horchanged) or (vertical_trigger and verchanged):
                    self.trigger_rotation = True

                print("Horizontal-vertical is {}".format(current_angle[0]))
                self.previous_angle = current_angle
                change = True
            else:
                break

        if not change:
            self.trigger_rotation = False

        # Run macro if set
        if self.macro:
                
            next_macro_step = False

            action = self.macro[self.i_macro]
            print(action)

            
            if type(action) == type((0,0)):
                # Move motors only if they have reached their positions
                if self.motors and all([self.motors[i].reached_target() for i in [0,1]]):
                    self.motors[0].move_to(action[0])
                    self.motors[1].move_to(action[1])
                    next_macro_step = True
            if 'wait' in action:
                self.waittime = time.time() + float(action.split(' ')[-1])
                next_macro_step = True
            
            if next_macro_step and self.waittime < time.time():
                self.i_macro += 1
                if self.i_macro == len(self.macro):
                    self.macro = None
                    self.i_macro = 0

    def set_zero(self):
        '''
        Define the current angle pair as the zero point
        (Like tar button on scales)
        '''
        self.reader.current_as_zero()



    def finalize(self):
        '''
        Finalising the experiments, houskeeping stuff.
        '''
        self.set_led(self.dynamic_parameters['ir_channel'], 0)
        self.set_led(self.dynamic_parameters['flash_channel'], 0)
        
        for motor in self.motors:
            motor.move_to(0)

        if self.triggered_anglepairs:
            fn = os.path.join(self.data_savedir, self.preparation['name'], 'anglepairs.txt')
            os.makedirs(os.path.dirname(fn), exist_ok=True)

            print(fn)
            
            with open(fn, 'w') as fp:
                for line in self.triggered_anglepairs:
                    fp.write(str(line).strip('()').replace(' ', '')+'\n')

            self.triggered_anglepairs = None


    def exit(self):
        for camera in self.cameras:
            camera.close_server()

    #
    # CONTROLLING LEDS, MOTORS ETC
    #
    
    def move_motor(*args, **kwargs):
        self.reader.move_motor(*args, **kwargs)


    #
    # PROTOCOL QUEUE / MACRO
    #


    @staticmethod
    def list_macros():
        return macro.list_macros()


    def run_macro(self, macro_name):
        self.macro = macro.load(macro_name)
        self.i_macro = 0







'''A terminal user interface library. Small, custom.
'''

import os
import platform
import time
import inspect      # Inspect docs and source code

OS = platform.system()
if OS == 'Windows':
    import msvcrt   # reading pressed keys without blocking
else:
    import sys
    import select


class SimpleTUI:
    '''A terminal user interface (TUI) library.
    
    Attributes
    ----------
    header : string or None
        Text shown on on the top of the screen.
    '''
    
    def __init__(self):
        self.header = None


    @staticmethod
    def read_key():
        if OS == 'Windows':
            if msvcrt.kbhit():
                key = ord(msvcrt.getwch())
                return chr(key)
        else:
            if select.select([sys.stdin], [], [], 0.1):
                return sys.stdin.read(1)
        return ''

    def clear_screen(self):
        '''Empties the screen and prints the header if any.
        '''
        if os.name == 'posix':
            os.system('clear')
        elif os.name == 'nt':
            os.system('cls')

        if self.header is not None:
            self.print(self.header)


    @staticmethod
    def print_lines(lines):
        for text in lines:
            print(text)
        
    @staticmethod
    def print(value=''):
        print(value)


    def _convert_selection(self, selection):
        '''
        selection : string
            Raw string typed by the user.
        '''
        selection = selection.strip('\r\n')
        selection = selection.split(',')

        # Raises ValueError if fails
        selection = [int(string) for string in selection]
        
        return selection


    def item_select(self, items, message=None, multiselect=False):
        '''Makes the user to select an item

        Arguments
        ---------
        items : iterable
            An iterable (eg. list) that returns printable items.
            If the item is newline, then prints a space and this
            space cannot be selected.
        message : string
            Printed before the selection table as an instruction
        multiselect : bool
            If True, 

        Returns
        -------
        item : ?
            The selected item. If multiselect is True, returns a
            list of selected items

        Empty string items are converted to a space
        '''
        
        if message is not None:
            self.print(message)

        real_items = []
        i = 0
        for item in items:
            if item != '\n':
                self.print('{}) {}'.format(i+1, item))
                real_items.append(item)
                i += 1
            else:
                self.print()

        self.print()
        
        if multiselect:
            self.print('# Multiple selections possible, separate by commas (,)')
        else:
            self.print('# Select by number and hit enter')

        self.print()

        selection = ''
        while True:
            new_char = self.read_key()
            if new_char:
                if new_char == '\b':
                    # Backspace
                    selection = selection[:-1]
                elif new_char in "0123456789\n\r":
                    selection += new_char
                self.print(selection)
            if selection.endswith('\r') or selection.endswith('\n'):
                
                try:
                    selection = self._convert_selection(selection)
                except ValueError:
                    self.print('Invalid input (non-numbers inputted)')
                    selection = ''
                    continue
                
                try:
                    # Test that we can get the item; -1 comes from
                    # the difference between Python first index (0)
                    # and this selection program first index (1) 
                    [real_items[number-1] for number in selection]
                except IndexError:
                    self.print('Invalid input (index out of range)')
                    selection = ''
                    continue

                break


        to_return = [real_items[int(number)-1] for number in selection]
        if not multiselect:
            to_return = to_return[0]

        return to_return

    def bool_select(self, message=None, true='yes', false='no'):
        '''Ask the user yes/no.

        Uses input so it blocks.
        '''
        
        if message is not None:
            self.print(message)

        while True:
            sel = input('(yes/no) >> ').lower()
            if sel.startswith('yes') or sel == 'y':

                return True
            elif sel.startswith('no') or sel == 'n':
                return False
            
            self.print('What? Please try again')
            time.sleep(1)



    def input(self, message=None, cancels=None):
        '''Ask the user for text input.

        A blocking call.
        
        Arguments
        ---------
        message : string
            The on-line message shown to the user.
        cancels: string
            If user inputs this string then returns None

        Returns
        -------
        user_input : string or None if cancels
        '''
        if message is not None:
            uinput = input(f'{message} >> ')
        else:
            uinput = input('>> ')
        
        if isinstance(cancels, str) and uinput == cancels:
            return None

        return uinput






'''
1) Reading rotation encoder signals from an Arduino Board.
2) and controlling stepper motors.
'''

try:
    import serial
    from serial.tools.list_ports import comports
except ModuleNotFoundError:
    serial = None



class ArduinoReader:
    '''
    Class for reading  angle pairs (states of the rotation encoders) from Arduino.
    '''

    def __init__(self, port=None):
        '''
        port        On Windows, "COM4" or similar. May change if other serial devices
                    are addded or removed?
        '''
        
        if serial:
            if port is None:
                # No port provided, use the first one with text "Arduino" in it
                ports = [str(p) for p in comports()]
                print(f'Selecting an Arduino from {ports}')
                for port in ports:
                    if not 'arduino' in port.lower():
                        continue
                    print(f'  Trying {port}')
                    try:
                        self.serial = serial.Serial(port=port.split(' ')[0], baudrate=9600, timeout=0.01)
                        self.serial.readline()
                        print(f'Accepted port {port}')
                        break
                    except Exception as e:
                        print(e)
            else:
                # Use the provided port
                self.serial = serial.Serial(port=port, baudrate=9600, timeout=0.01)
        else:
            self.serial = None

        self.latest_angle = (0,0)
        self.offset = (0,0)

    def _offset_correct(self, angles):
        '''
        Rreturn the offset (zero-point) corrected angles pair.
        '''
        return (angles[0] - self.offset[0], angles[1] - self.offset[1])


    def read_angles(self):
        '''
        Read the oldest unread angles pair that Arduino has sent to the serial.

        Returns angle pair, (horizontal_angle, vertical_angle).
        '''
        if self.serial is None:
            return (0,0)

        read_string = self.serial.readline().decode("utf-8")
        if read_string:
            angles = read_string.split(',')
            self.latest_angle = tuple(map(int, angles))

        return self._offset_correct(self.latest_angle)

    def get_latest(self):
        '''
        Returns the latest angle that has been read from Arduino.
        (Arduino sends an angle only when it has changed)
        '''
        return self._offset_correct(self.latest_angle)
    

    def close_connection(self):
        '''
        If it is required to manually close the serial connection.
        '''
        if self.serial:
            self.serial.close()

    def current_as_zero(self):
        '''
        Sets the current angle pair value to (0,0)
        '''
        self.offset = self.latest_angle

    
    def move_motor(self, i_motor, direction, time=1):
        '''
        Move motor i_motor to given direction for the given time (default 1 s)

        Communication to the Arduino board controlling the motor states
        happens by sending characters to the serial; each sent character makes
        the motor to move for 100 ms ideally; letters as
            a       for the motor 0 to + direction
            A       for the motor 0 to - direction
            b       for the motor 1 to + direction
            ....

        
        Input arguments
        i_motor         Index number of the motor
        direction       Positive for + direction, negative number for - direction
                        0 makes nothing
        time            In seconds
        '''
        if self.serial is None:
            print('Pretending to drive motor {}'.format(i_motor))
            return None

        motor_letters = ['a', 'b', 'c', 'd', 'e']
    
        letter = motor_letters[i_motor]
        
        if not direction == 0:
            

            if direction > 0:
                letter = letter.lower()
            else:
                letter = letter.upper()
            
            N = round(time * 10)
            string = ''.join([letter for i in range(N)])

            self.serial.write(bytearray(string.encode()))

    def get_sensor(self, i_sensor):
        '''
        Yet another way to read anglepairs, separated.
        '''
        angles = self.get_latest()
        return angles[i_sensor]




'''Shared code between various servers.
'''

import socket
import time
import os

DEFAULT_SAVE_DIRECTORY = 'gonioimsoft_data'


class ServerBase:
    '''The base class for any server.

    Attributes
    ----------
    socket : obj
        The socket object.
    device : obj or None
        
    functions : dict
        Keys command names and values callables.
    responders : list
        Command names (a subset from functions) that also need to
        send a return value to the client.
    '''
    
    def __init__(self, host, port, device):
        
        print(f'Binding a socket (host {host}, port {port}')
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socket.bind(('', port))
        self.socket.listen(1)
        
        if getattr(device, 'save_directory', None):
            device.save_directory = DEFAULT_SAVE_DIRECTORY
        self.device = device

        
        self.functions = {
                'ping': self.ping,
                'pong': self.pong,
                'exit': self.exit,
                'set_save_directory': self.set_save_directory,
                }

        self.responders = ['pong']

        self.run_exit = False



    def ping(self, message):
        '''Prints the client's message on the server.
        '''
        print(f'A client says: {message}')


    def pong(self, message):
        '''Sends the client's message back to the server with a greeting.
        '''
        return f'General Kenobi! (response to {message})'
    
    def set_save_directory(self, directory):
        '''Sets the location for the data saving
        '''
        abspath = os.path.abspath(directory)
        if not os.path.isdir(directory):
            print(f'Creating directory {abspath} for saving data')
            os.makedirs(directory)
        else:
            print(f'Setting directory {abspath} for saving data')
        
        if self.device is not None:
            self.device.save_directory = directory


    def wait_for_client(self):
        '''Waits for any command from the client and then discards it.
        '''
        print('Waiting for any command...')
        conn, addr = self.socket.accept()
        while True:
            data = conn.recv(1024)
            if not data: break
        conn.close()
        print('Got a command from the client, waiting done!')


    def run(self):
        '''Runs the server mainloop until receives an exit command.

        In each turn of the loop, waits for incoming connection and
        executes the clients wishes.
        '''
        print('Waiting clients to connect')
        while not self.run_exit:
            conn, addr = self.socket.accept()
            data = conn.recv(1024)
            string = data.decode()

            if not string:
                conn.close()
                continue

            print(f'Got command {string} at {time.time()}')

            # The char ';' is used as a delimiter between the
            # command name and the arguments. Parameters by ':'
            if ';' in string:
                func, parameters = string.split(';')
                parameters = parameters.split(':')
            else:
                func = string
                parameters = []

            if func not in self.functions:
                print(f'Skipping unkown command: {func}')
                conn.close()
                continue

            # If the client expects no response, then we can close the
            # connection early and let the client go.
            if not func in self.responders:
                conn.close()
            
            try:
                response = self.functions[func](*parameters)
            except Exception as e:
                print()
                print('Failure running the command')
                print(f'  Invocation (Python): {func}(*{parameters})')
                print('  Error below:')
                print(e)
                print()
                response = 'error'

                
            # Say back the response and close
            if func in self.responders:

                if isinstance(response, (list, tuple)):
                    response = ':'.join(response)

                conn.sendall(str(response).encode())
                conn.close()


    def exit(self, _=None):
        '''Makes the run function to exit from its mainloop.
        '''
        self.run_exit = True



'''Settting and getting imaging parameters.

Default imaging parameters, and ParameterEditor for letting the user
modify them in the program.
'''

import os
import time
import ast
import json

from gonioimsoft.directories import USERDATA_DIR


DEFAULT_DYNAMIC_PARAMETERS = {
        'isi': 10.0, 'repeats': 1, 'pre_stim': 0.000,
        'stim': 0.200, 'post_stim': 0.00, 'frame_length' : 0.010,
        'ir_imaging': 5, 'ir_waiting': 0, 'ir_livefeed': 1,
        'flash_on': 5, 'flash_off': 0,
        'ir_channel': "Dev1/ao1", 'flash_channel': "Dev1/ao0",
        'suffix': '', 'trigger_channel': "/Dev1/PFI0",
        'trigger_out_channel': "Dev1/ao7",
        'biosyst_stimulus': '',
        'biosyst_channel': 2,
        'avgint_adaptation': 0,
        'flash_type': 'square',
        'save_stack': True,
        'reboot_cameras': False,
        'ROI': None,
        }

DYNAMIC_PARAMETERS_TYPES = {
        'seconds': ['isi', 'pre_stim', 'stim', 'post_stim', 'frame_length', 'avgint_adaptation'],
        'voltage': ['ir_imaging', 'ir_waiting', 'ir_livefeed', 'flash_on', 'flash_off'],
        'channel': ['ir_channel', 'flash_channel', 'trigger_channel', 'trigger_out_channel'],
        'integer': ['repeats', 'biosyst_channel'],
        'string': ['suffix', 'biosyst_stimulus', 'flash_type'],
        'boolean': ['save_stack', 'reboot_cameras'],
        'roibox': ['ROI']}


DYNAMIC_PARAMETERS_HELP = {
        'isi': 'Inter stimulus interval [s]',
        'repeats': 'How many times the protocol is repeated [1-inf]',
        'pre_stim': 'How long to image before the pulse [s]',
        'stim': 'Stimulus (step pulse) length [s]',
        'post_stim': 'How long to image after the pulse [s]',
        'frame_length': 'Exposure time / inter-frame interval [s]',
        'ir_imaging': 'IR brightness during image acqusition [0-10]',
        'ir_waiting': 'IR brightness when waiting ISI [0-10]',
        'ir_livefeed': 'IR brightness while updating the live image[0-10]',
        'flash_on': 'Flash brightness during stim [0-10]',
        'flash_off': 'Flash brightness during pre- and post-stim [0-10]',
        'ir_channel': 'NI channel for IR [0-10]',
        'flash_channel': 'NI channel for Flash',
        'trigger_channel': 'Trigger recieve/in channel for NI',
        'trigger_out_channel': 'Trigger send/out channel from NI',
        'suffix': 'Tag added to the saved folders',
        'biosyst_stimulus': 'Override the square pulse by a biosyst stimulus [filename]',
        'biosyst_channel': 'The channel read from the biosyst simulus file if set',
        'avgint_adaptation': 'Time to show stimulus mean value before imaging [s]',
        'flash_type': 'square, sinelogsweep, squarelogsweep or 3steplogsweep. "{sweep},f0,f1" for Hz',
        'save_stack': 'If true, save a stack instead separate images',
        'reboot_cameras': 'If true, reboots cameras after each run (dirtyfix)',
        'ROI': 'If set, crops the sensor area (allows higher fps). x,y,w,h',
        }


def getRightType(parameter_name, string_value):
    '''
    Convert user inputted string to correct parameter value based on
    DYNAMIC_PARAMETER_TYPES

    TODO    - channel checking, check that the channel is proper NI channel
    '''
    if parameter_name in DYNAMIC_PARAMETERS_TYPES['integer']:
        return int(string_value)

   
    if parameter_name in DYNAMIC_PARAMETERS_TYPES['seconds']:
        if string_value.startswith('[') and string_value.endswith(']'):
            seconds = ast.literal_eval(string_value)

            for s in seconds:
                if s < 0:
                    raise ValueError('Here time is required to be strictly positive.')
        else:
            seconds = float(string_value)
            if seconds < 0:
                raise ValueError('Here time is required to be strictly positive.')
        return seconds

    if parameter_name in  DYNAMIC_PARAMETERS_TYPES['voltage']:
        if string_value.startswith('[') and string_value.endswith(']'):
            voltages = ast.literal_eval(string_value)
            for voltage in voltages:
                if not -10<=voltage<=10:
                    raise ValueError('Voltage value range -10 to 10 V exceeded.')
            return voltages
        else:
            voltage = float(string_value)
            if not -10<=voltage<=10:
                raise ValueError('Voltage value range -10 to 10 V exceeded.')
            return voltage

    if parameter_name in  DYNAMIC_PARAMETERS_TYPES['channel']:
        if type(string_value) == type(''):
            if string_value.startswith('[') and string_value.endswith(']'):
                return ast.literal_eval(string_value)
            else:
                return string_value
    
    if parameter_name in DYNAMIC_PARAMETERS_TYPES['string']:
        return str(string_value)

    if parameter_name in DYNAMIC_PARAMETERS_TYPES['boolean']:
        if string_value.lower() == 'true':
            return True
        elif string_value.lower() == 'false':
            return False
        else:
            raise ValueError('Boolean falue has to be either "True" or "False"')


    if parameter_name in DYNAMIC_PARAMETERS_TYPES['roibox']:
        try:
            x,y,w,h = [int(num) for num in string_value.split(',')]
            return (x,y,w,h)
        except:
            return None

    raise NotImplementedError('Add {} correctly to DYNAMIC_PARAMETER_TYPES in dynamic_parameters.py')



def load_parameters(fn):
    '''
    Loading imaging parameters, saved as a json file.
    '''
    with open(fn, 'r') as fp:
        data = json.load(fp)
    return data


def save_parameters(fn, parameters):
    '''
    Loading imaging parameters, saved as a json file.
    '''
    with open(fn, 'w') as fp:
        json.dump(parameters, fp)



class ParameterEditor:
    '''
    Dictionary editor on command line with ability to load and save presets.
    '''
    def __init__(self, dynamic_parameters):
        '''
        dynamic_parameters      Dictionary of the dynamic imaging parameters.
        '''
        self.dynamic_parameters = dynamic_parameters
        self.parameter_names = sorted(self.dynamic_parameters.keys())

        self.presets_savedir = os.path.join(USERDATA_DIR, 'presets')
        self.presets = self.load_presets(self.presets_savedir)

    
    def load_presets(self, directory):
        
        presets = {}
        
        if os.path.isdir(directory):
            files = [os.path.join(directory, fn) for fn in os.listdir(directory)]
        else:
            files = []

        for afile in files:
            try:
                preset = load_parameters(afile)
            except:
                print("Couldn't load preset {}".format(afile))
                continue
            
            # If older files lack some parameters, use default parameters
            for key in self.parameter_names:
                if not key in preset:
                    preset[key] = DEFAULT_DYNAMIC_PARAMETERS[key]
                    
            presets[os.path.basename(afile)] = preset

        return presets


    def print_preset(self, preset):
        '''
        Prints the current parameters and the help strings.
        '''
                 
        parameter_names = sorted(self.dynamic_parameters.keys())

        print()
        
        print('{:<20} {:<40} {}'.format('PARAMETER NAME', 'VALUE', 'DESCRIPTION'))
        for parameter in parameter_names:

            value = str(preset[parameter])

            # Special addition for frame_length parameter; Show Hz so that
            # it is harder to do mistakes
            if parameter == 'frame_length':
                hz = str(round(1/float(value)))
                if hz == 1:
                    hz = str(1/float(value))
                value += ' ({} Hz)'.format(hz)

            print('{:<20} {:<40} {}'.format(
                parameter, value, DYNAMIC_PARAMETERS_HELP[parameter]))
        print()


    def getModified(self):
        '''
        Ask user to edit the parameters until happy and then return
        the parameters.
        '''
        
        while True:
            print('MODIFYING IMAGING PARAMETERS')
            self.print_preset(self.dynamic_parameters)
            parameter = input('Parameter name or (list/save/load/back) (Enter to continue) >> ')
            
            # If breaking free
            if parameter == '':
                break
            

            self.presets = self.load_presets(self.presets_savedir)

            
            # If saving preset
            if parameter.lower() == 'save':
                name = input('Save current parameters under preset name (if empty == suffix) >> ')
                if name == '' and self.dynamic_parameters['suffix'] != '':
                    name = self.dynamic_parameters['suffix']
                
                if os.path.isdir(USERDATA_DIR):
                    os.makedirs(self.presets_savedir, exist_ok=True)
                    save_parameters(os.path.join(self.presets_savedir, name), self.dynamic_parameters)
                else:
                    print('Saving the preset failed, {} does not exist'.format(USERDATA_DIR))

                continue        

            if parameter.lower() == 'list':
                if self.presets is {}:
                    print('There are no existing presets!')
                else:
                    print('These are the existing presets:')
                    for preset in self.presets.keys():
                        print('  '+preset)
                    print('')
                continue

            if parameter.lower() == 'load':
                # If parameter is actually a preset
                
                while True:
                    
                    preset_names = sorted(self.presets.keys())

                    for i, name in enumerate(preset_names):
                        print('{}) {}'.format(i+1, name))
                    
                    sel = input('>> ')

                    try:
                        to_load = preset_names[int(sel)-1]
                        break
                    except:
                        print('Invalid preset.')

                parameter = to_load

            if parameter == 'back':
                return None

            if parameter in self.presets.keys():
                self.print_preset(self.presets[parameter])
                
                confirm = input('Load this (y/n)>> ').lower()
                if confirm and confirm[0] == 'y':
                    self.dynamic_parameters = self.presets[parameter]
                else:
                    print('Answer not yes, loading of the preset cancelled')
                continue
            
    

            try:
                self.dynamic_parameters[parameter]
            except KeyError:
                print('Invalid input, not a parameter name')
                time.sleep(1)
                continue

            while True:

                value = input('Value for {} >> '.format(parameter))
                
                if value == '':
                    break

                try:
                    value = getRightType(parameter, value)
                except ValueError as e:
                    print(str(e))
                    print('Could not convert the value to right type. Try againg or Enter to skip.')
                    continue

                self.dynamic_parameters[parameter] = value
                break

        return self.dynamic_parameters


def getModifiedParameters(**kwargs):
    '''
    Take in the DEFAULT parameters in the beginning of this code file
    and let the user modify them using text based ParameterEditor
    '''
    editor = ParameterEditor(DEFAULT_DYNAMIC_PARAMETERS, **kwargs)
    return editor.getModified()



if __name__ == "__main__":
    print(getModifiedParameters())





'''Common settings shared by the camera server/client
'''


# Hostname (or IP address) of the server. This is the address that the
# client tries to connect to
SERVER_HOSTNAME = '127.0.0.1'
CAMERA_PORT = 50075

# Voltage input output server/client
VIO_PORT = 50085




'''Voltage input/output server.
'''

import os
import multiprocessing

import numpy as np
import matplotlib.pyplot as plt

try:
    import nidaqmx
except ImportError:
    nidaqmx = None
    print('nidaqmx not available')

from .common import VIO_PORT
from .serverbase import ServerBase


class Plotter:
         
    def loop(self, queue, title):
        
        fig = plt.figure()
        ax = fig.add_subplot(111)
        plt.show(block=False)

        fig.canvas.toolbar.winfo_toplevel().title(title)
        
        while True:
            if queue.empty():
                plt.pause(0.1)
                continue
            while not queue.empty():
                data = queue.get()
            if isinstance(data, str) and data == 'close':
                break

            ax.clear()
            ax.plot(data)
            plt.pause(0.1)
        


def run_plotter(queue, title):
    plotter = Plotter()
    plotter.loop(queue, title)


class DummyBoard:
    '''For testing the server without an NI board.

    For documentation, see the NIBoard class.
    '''
    def analog_input(self, duration, wait_trigger=False):
        print('DummyBoard.analog_input(...)')
        print(f'dur={duration} | wait_trigger={wait_trigger}')

    def set_settings(self, device, channels, fs):
        print('DummyBoard.set_settings(...)')
        print(f'dvs={device} | chs={channels} | fs={fs}')

class NIBoard:
    def __init__(self):

        self.device = 'Dev2'
        self.channels = ['ai0']
        self.fs = 1000

        self.live_queue = None
        self.title = 'Analog input'
    

    def set_settings(self, device, channels, fs):
        '''

        Arguments
        ---------
        dev : string
            The NI board name (usually "Dev1" or "Dev2").
        channels : string
            Channels names to record, separated by commas.
        fs : float or int
            The used sampling frequency in Hz (samples/second).
        '''
        self.device = device
        self.channels = channels.split(',')
        self.fs = float(fs)

        self.title = f'Analog input - {self.device} {self.channels} {self.fs} Hz'
        
   
    def analog_input(self, duration, save=None, wait_trigger=False):
        '''Records voltage input and saves it.

           
        duration : int or float
            In seconds, the recording's length.

        '''
        if save == 'None':
            save = None
            
        duration = float(duration)
        timeout = duration + 10

        N_channels = len(self.channels)
        N_samples = int(duration * self.fs)
        

        if str(wait_trigger).lower() == 'true':
            wait_trigger = True
        else:
            wait_trigger = False


        with nidaqmx.Task() as task:
            
            for channel in self.channels:
                task.ai_channels.add_ai_voltage_chan(f'{self.device}/{channel}')

            task.timing.cfg_samp_clk_timing(
                    self.fs, samps_per_chan=N_samples)
            task.ai_conv_rate = self.fs
            if wait_trigger:
                task.triggers.start_trigger.cfg_dig_edge_start_trig(
                        f'/{self.device}/PFI0')

            task.start()

            data = task.read(
                timeout=timeout, number_of_samples_per_channel=N_samples)


        if save is not None:
            fn = os.path.join(
                self.save_directory, f'{save}.npy')
            np.save(fn, np.array(data))


        if self.live_queue is None:
            self.live_queue = multiprocessing.Queue()
            self.live_queue.put(data)
            
            self.livep = multiprocessing.Process(
                    target=run_plotter,
                    args=(self.live_queue,self.title))
            self.livep.start()
            
        else:
            self.live_queue.put(data)


class VIOServer(ServerBase):
    '''Analog voltage input/output board server.
    '''

    def __init__(self, device, port=None):

        if port is None:
            port = VIO_PORT
        super().__init__('', port, device)
        
        self.functions['analog_input'] = self.device.analog_input
        self.functions['set_settings'] = self.device.set_settings



def main():
        
    if nidaqmx is None:
        board = DummyBoard()
    else:
        board = NIBoard()

    server = VIOServer(board)
    server.run()


if __name__ == "__main__":
    main()



import numpy as np

import macro


def main():
    horizontals = [-50, -40, -30, -20, -10, 0, 10, 20, 30]
    verticals = np.arange(-20, 160, 10)

    new_macro = []

    for i, vertical in enumerate(verticals):
        horizontals.reverse()
            
        for horizontal in horizontals:
            new_macro.append((horizontal, vertical))
            new_macro.append('wait 1')

    macro.save('default', new_macro)

if __name__ == "__main__":
    main()



'''
Working with anglepairs.txt file
'''

import csv


def saveAnglePairs(fn, angles):
    '''
    Saving angle pairs to a file.
    '''
    with open(fn, 'w') as fp:
        writer = csv.writer(fp)
        for angle in angles:
            writer.writerow(angle)

def loadAnglePairs(fn):
    '''
    Loading angle pairs from a file.
    '''
    angles = []
    with open(fn, 'r') as fp:
        reader = csv.reader(fp)
        for row in reader:
            if row:
                angles.append([int(a) for a in row])
    return angles

def toDegrees(angles):
    '''
    Transform 'angles' (that here are just the steps of rotary encoder)
    to actual degree angle values.
    '''
    for i in range(len(angles)):
        angles[i][0] *= (360/1024)
        angles[i][1] *= (360/1024)


def step2degree(step):
    return step * (360/1024)


def degrees2steps(angle):
    '''
    Transform an angle (degrees) to corresponging rotary encoder steps.
    '''
    return angle*(1024/360)




'''The camera server for using the camera and save images.

Functionality
-------------
- Image acquisition using Micro-Manager's Python bindings (pymmcore)
- Live feed using matplotlib
- File saving using the tifffile module


Background
-----------
On Windows, MM builds came precompiled with Python 2 support only. The
camera server/client division here made it possible to run the server
calling MM using Python 2 (to control the camera and image saving) and
then the camera client run with Python 3.

It is still usefull; Because it uses (network) sockets, it is trivial
to use another (or multiple PC) to acqurie images. Or, in future, to
use another backends than MM.
'''

import os
import sys
import time
import datetime
import argparse
import threading
import multiprocessing

try:
    import pymmcore
except ImportError:
    pymmcore = None
    print('pymmcore not installed')
import tifffile
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.animation import FuncAnimation
from matplotlib.widgets import RectangleSelector

from .common import CAMERA_PORT
from .serverbase import ServerBase

DEFAULT_MICROMANAGER_DIR = 'C:/Program Files/Micro-Manager-2.0'

# Integer between 1-inf (1 = no downsampling), images for imageshower
LIVE_DOWNSAMPLE = 2

class ImageShower:
    '''Shows images on the screen in its own window.

    In future, may be used to select ROIs as well to allow
    higher frame rate imaging / less data.
    
    ------------------
    Working principle
    ------------------
    Image shower works so that self.loop is started as a separate process
    using multiprocessing library
    
    -------
    Methods
    -------
    self.loop       Set this as multiprocessing target
    '''
    def __init__(self):
        self.fig = plt.figure()
        self.ax = self.fig.add_subplot(111)
        self.close = False

        #self.cid = self.fig.canvas.mpl_connect('key_press_event', self.callbackButtonPressed)
        
        self.image_brightness = 0
        self.image_maxval = 1

        self.selection = None

        self.image_size = None

    def callbackButtonPressed(self, event):
        
        if event.key == 'r':
            self.image_maxval -= 0.05
            self._updateImage(strong=True)
        
        elif event.key == 't':
            self.image_maxval += 0.05
            self._updateImage(strong=True)

            

    def __onSelectRectangle(self, eclick, erelease):
        
        # Get selection box coordinates and set the box inactive
        x1, y1 = eclick.xdata, eclick.ydata
        x2, y2 = erelease.xdata, erelease.ydata
        #self.rectangle.set_active(False)
        
        x = int(min((x1, x2)))
        y = int(min((y1, y2)))
        width = int(abs(x2-x1))
        height = int(abs(y2-y1))
        
        self.selection = [x, y, width, height]
        
    def _updateImage(self, i):
        
        data = None
        while not self.queue.empty():
            # Get the latest image in the queue
            data = self.queue.get(True, timeout=0.01)
        if data is None:
            return self.im, ''
        elif isinstance(data, str) and data == 'close':
            self.close = True
            return self.im, ''

        if self.selection and data.size != self.image_size:
            self.selection = None

        if self.selection:
            x,y,w,h = self.selection
            if w<1 or h<1:
                # If selection box empty (accidental click on the image)
                # use the whole image instead
                inspect_area = data
            else:
                inspect_area = data[y:y+h, x:x+w]
        else:
            inspect_area = data
        
        
        per95 = np.percentile(inspect_area, 95)
        per5 = np.percentile(inspect_area, 5)
        data = np.clip(data, per5, per95)
        
        data -= per5
        data /= (per95-per5)

        self.image_size = data.size
        
        self.im.set_array(data)
        self.fig.suptitle('Selection 95th percentile: {}'.format(per95), fontsize=10)
        text = ''
        return self.im, text
           
         
    def loop(self, queue, title):
        '''
        Runs the ImageShower by reading images from the given queue.
        Set this as a multiprocessing target.

        queue           Multiprocessing queue with a get method.
        '''
        self.queue = queue
        self.rectangle = RectangleSelector(self.ax, self.__onSelectRectangle, useblit=True)
        
        image = queue.get()
        self.im = plt.imshow(1000*image/np.max(image), cmap='gray', vmin=0, vmax=1, interpolation='none', aspect='auto')
        self.ani = FuncAnimation(plt.gcf(), self._updateImage, frames=range(100), interval=50, blit=False)

        self.fig.canvas.toolbar.winfo_toplevel().title(title)

        # Remove the toolbar; Gives more space when having many cameras
        self.fig.canvas.toolbar.pack_forget()
        
        # Take away all white space
        plt.subplots_adjust(top=1, bottom=0, right=1,left=0,
                            hspace=0, wspace=0)
        
        plt.show(block=True)


class DummyCamera:
    '''A dummy camera suitable for testing the server/client.
    '''
    def __init__(self):
        self.settings = {'setting1' : 'na', 'setting2': 0.0, 'setting3': 1}
        self.camera = None

    def acquire_single(self, save, subdir):
        pass
    def acquire_series(self, exposure_time, image_interval, N_frames, label, subdir):
        pass
    def save_images(images, label, metadata, savedir):
        pass
    def set_binning(self, binning):
        self.settings['binning'] = binning
    def set_roi(self, x,y,w,h):
        self.settings['roi'] = [x,y,w,h]
    def set_save_stack(self, boolean):
        self.settings['save-stack'] = boolean

    def save_description(self, filename, string):
        pass
    def close(self):
        pass
    def get_cameras(self):
        return ['dummy1', 'dummy2']
    def get_camera(self):
        return self.camera
    def set_camera(self, name):
        self.camera = name
    def get_settings(self):
        return list(self.settings.keys())
    def get_setting_type(self, setting_name):
        if setting_name == 'setting1':
            return 'string'
        elif setting_name == 'setting2':
            return 'float'
        elif setting_name == 'setting3':
            return 'integer'
        else:
            print('Invalid setting')
            return ''
    def get_setting(self, setting_name):
        value = self.settings.get(setting_name, None)
        if value is None:
            return ''
        return value
    def set_setting(self, setting_name, value):
        self.settings[setting_name] = value



class MMCamera:
    '''Controls any camera using MicroManager and its pymmcore bindings.
    '''

    def __init__(self):

        self.mmc = pymmcore.CMMCore() 
        self.mmc.setDeviceAdapterSearchPaths([DEFAULT_MICROMANAGER_DIR])

        self._device_name = None
        self._configuration_name = ''

        #self.mmc.loadDevice('Camera', 'HamamatsuHam', 'HamamatsuHam_DCAM')
        #self.mmc.initializeAllDevices()
        #self.mmc.setCameraDevice('Camera')
            
        self.settings = {'exposure_time_scaler': 1}
        
        #self.mmc.setCircularBufferMemoryFootprint(4000)
        self.live_queue= False

        self.shower = ImageShower()

        # Description file string
        self.description_string = ''

        self.save_stack = False
        self.save_directory = None

        self._startdir = os.getcwd()

        self.title = 'Camera not set'
        self.servertitle = ''


    def get_cameras(self):
        '''Lists available MicroManager configuration files in .
        '''
        return [fn for fn in os.listdir(DEFAULT_MICROMANAGER_DIR) if fn.endswith('.cfg')]


    def get_camera(self):
        '''Returns the label of the current camera.
        '''
        return self._configuration_name

    def set_camera(self, name):
        '''Set the provided configuration file.
        '''
        if not os.path.exists(name):
            path = os.path.join(DEFAULT_MICROMANAGER_DIR, name)
            
            if not os.path.exists(path):
                print(f'Couldnt open configuration file named: {path}')
                return

        self.mmc.loadSystemConfiguration(path)
        
        self._device_name = self.mmc.getCameraDevice()
        self._configuration_name = name
        self.mmc.prepareSequenceAcquisition(self._device_name)

        self.title = f'{name} ({self._device_name}) | {self.servertitle}'

    def get_settings(self):
        '''Returns device property names
        '''
        if self._device_name is None:
            return ''
        properties = self.mmc.getDevicePropertyNames(self._device_name)
        
        return list(properties) + list(self.settings.keys())


    def get_setting_type(self, setting_name):
        '''Returns "string", "integer" or "float".

        Returns an empty string if the setting does not exist.
        '''
        if setting_name in self.settings:
            return 'float'

        try:
            num = self.mmc.getPropertyType(self._device_name, setting_name)
        except RuntimeError as e:
            print(f'Error! No setting named: {setting_name}')
            return ''
        
        if num == 1:
            return 'string'
        elif num == 2:
            return 'float'
        elif num == 3:
            return 'integer'


    def get_setting(self, setting_name):
        if setting_name in self.settings:
            return self.settings[setting_name]
        return self.mmc.getProperty(self._device_name, setting_name)

    def set_setting(self, setting_name, value):
        
        # a) Internal setting
        if setting_name in self.settings:
            self.settings[setting_name] = value
            return
        
        # b) Camera device setting
        type_name = self.get_setting_type(setting_name)
        if type_name == 'float':
            value = float(value)
        elif type_name == 'integer':
            value = int(value)
        elif type_name == '':
            print('Error! No setting named: {setting_name}')
            return

        print(f'Changing {setting_name} to its new value {value}')
        try:
            self.mmc.setProperty(self._device_name, setting_name, value)
        except RuntimeError as e:
            print('Error! The set value likely out of range.')
            print(e)

    def acquire_single(self, exposure_time, save, subdir):
        '''
        Acquire a single image.

        save        'True' or 'False'
        subdir      Subdirectory for saving
        '''
        
        exposure_time = float(exposure_time)
        self.mmc.setExposure(exposure_time*1000)
        #binning = '2x2'

        #self.set_binning(binning)

        start_time = str(datetime.datetime.now())
 
        self.mmc.snapImage()
        image = self.mmc.getImage()
        
        if not self.live_queue:
            self.live_queue = multiprocessing.Queue()
            self.live_queue.put(
                    image[0::LIVE_DOWNSAMPLE, 0::LIVE_DOWNSAMPLE])
            
            self.livep = multiprocessing.Process(
                    target=self.shower.loop,
                    args=(self.live_queue,self.title))
            self.livep.start()
            
        self.live_queue.put(image[0::LIVE_DOWNSAMPLE, 0::LIVE_DOWNSAMPLE])

        if save == 'True':
            metadata = {'exposure_time_s': exposure_time, 'function': 'acquireSingle', 'start_time': start_time}

            save_thread = threading.Thread(target=self.save_images,args=([image],'snap_{}'.format(start_time.replace(':','.').replace(' ','_')), metadata,os.path.join(self.save_directory, subdir)))
            save_thread.start()



    def acquire_series(self, exposure_time, image_interval, N_frames, label, subdir):
        '''
        Acquire a series of images

        exposure_time       How many seconds to expose each image
        image_interval      How many seconds to wait in between the exposures
        N_frames            How many images to take
        label               Label for saving the images (part of the filename later)
        subdir
        '''

        exposure_time = float(exposure_time)
        image_interval = float(image_interval)
        N_frames = int(N_frames)
        label = str(label)

        print("Now aquire_series with label " + label)
        print("- IMAGING PARAMETERS -")
        print(" exposure time " + str(exposure_time) + " seconds")
        print(" image interval " + str(image_interval) + " seconds")
        print(" N_frames " + str(N_frames))
        print("- CAMERA SETTINGS")

        #self.set_binning('2x2')
        #print(" Pixel binning 2x2")
        
        device_name = self.mmc.getDeviceName(self._device_name)
        print(f'Device name {device_name}')

        #if 'hamamatsu' in device_name.lower():
        #    if trigger_direction == 'send':
        #        print(" Camera sending a trigger pulse")
        #        self.mmc.setProperty(self._device_name, "OUTPUT TRIGGER KIND[0]","EXPOSURE")
        #        self.mmc.setProperty(self._device_name, "OUTPUT TRIGGER POLARITY[0]","NEGATIVE")
        #    elif trigger_direction== 'receive':
        #        print(" Camera recieving / waiting for a trigger pulse")
        #        self.mmc.setProperty(self._device_name, "TRIGGER SOURCE","EXTERNAL")
        #        self.mmc.setProperty(self._device_name, "TriggerPolarity","POSITIVE")
        #    elif trigger_direction == 'none':
        #        pass
        #    else:
        #        raise ValueError('trigger_direction has to be send, receive or none, not {trigger_direction}')

        
        print("Circular buffer " + str(self.mmc.getCircularBufferMemoryFootprint()) + " MB")

        scaler = float(self.settings['exposure_time_scaler'])
        exposure = scaler*exposure_time*1000
        self.mmc.setExposure(exposure)

        self.mmc.clearCircularBuffer()
        #self.mmc.prepareSequenceAcquisition(self._device_name)
        self.wait_for_client()
        
        start_time = str(datetime.datetime.now())
        self.mmc.startSequenceAcquisition(N_frames, image_interval+(1-scaler)*exposure, False)
        
        while self.mmc.isSequenceRunning():
            self.mmc.sleep(1000*exposure_time)

        self.mmc.sleep(1000)

        images = []

        for i in range(N_frames):
            while True:
                try:
                    image = self.mmc.popNextImage()
                    break
                except:
                    # Index error for example when circular buffer is still empty
                    self.mmc.sleep(1000*exposure_time)
            
            images.append(image)
            
            
        metadata = {'exposure_time_s': exposure_time, 'image_interval_s': image_interval,
                    'N_frames': N_frames, 'label': label, 'function': 'acquireSeries', 'start_time': start_time}
        metadata.update(self.settings)

        save_thread = threading.Thread(target=self.save_images, args=(images,label,metadata,os.path.join(self.save_directory, subdir)))
        save_thread.start()
        
        #if 'hamamatsu' in device_name.lower() and trigger_direction == 'receive':
        #    self.mmc.setProperty(self._device_name, "TRIGGER SOURCE","INTERNAL")
        
        print('acquired')

    
    def save_images(self, images, label, metadata, savedir):
        '''
        Save given images as grayscale tiff images.
        '''
        savedir = os.path.join(self._startdir, savedir)
        if not os.path.isdir(savedir):
            try:
                os.makedirs(savedir)
            except:
                # May fail if many local servers creating
                # the folders simultaneously
                pass

        if self.save_stack == False:
            # Save separate images
            for i, image in enumerate(images):
                fn = '{}_{}.tiff'.format(label, i)
                tifffile.imwrite(os.path.join(savedir, fn), image, metadata=metadata)
        else:
            # Save a stack
            fn = '{}_stack.tiff'.format(label)
            tifffile.imwrite(os.path.join(savedir, fn), np.asarray(images), metadata=metadata)
        
        self.save_description(os.path.join(savedir, 'description'), self.description_string, internal=True)


    def set_save_stack(self, boolean):
        '''
        If boolean == "True", save images as stacks instead of separate images.
        '''
        if boolean == 'True':
            self.save_stack = True
        elif boolean == 'False':
            self.save_stack = False
        else:
            print("Did not understand wheter to save stacks. Given {}".format(boolean))

    def set_binning(self, binning):
        '''
        Binning '2x2' for example.
        '''
        if not self.settings['binning'] == binning:
            self.mmc.setProperty(self._device_name, 'Binning', binning)
            self.settings['binning'] =  binning

    def set_roi(self, x,y,w,h):
        '''
        In binned pixels
        roi     (x,y,w,h) or None
        '''
        x = int(x)
        y = int(y)
        w = int(w)
        h = int(h)
        
        if w == 0 or h==0:
            self.mmc.clearROI()
        else:
            self.mmc.setROI(x,y,w,h)


    def save_description(self, specimen_name, desc_string, internal=False):
        '''
        Allows saving a small descriptive text file into the main saving directory.
        Filename should be the same as the folder where it's saved.

        Appends to the previous file.

        specimen_name           DrosoM42 for example, name of the specimen folder
        desc_string             String, what to write in the file
        internal                If true, specimen_name becomes filename of the file
        '''
        if internal:
            fn = specimen_name
        else:
            fn = os.path.join(self.save_directory, specimen_name, specimen_name)
        
        # Check if the folder exists
        if not os.path.exists(os.path.dirname(fn)):
            try:
                os.makedirs(os.path.dirname(fn), exist_ok=True)
            except:
                # My fail if many local servers. Let's not
                # care about it, another server has made it
                pass
        
        try:
            with open(fn+'.txt', 'w') as fp:
                fp.write(desc_string)

            print("Wrote file " + fn+'.txt') 
        except:
            pass
        
        self.description_string = desc_string


    def close(self):
        if self.live_queue:
            self.live_queue.put('close')

    def wait_for_client(self):
        pass


class CameraServer(ServerBase):
    '''Camera server listens incoming connections from the client and
    controls a camera class.
    '''

    def __init__(self, camera, port=None):
        
        if port is None:
            port = CAMERA_PORT
        
        super().__init__('', port, camera)
        
        
        print(f'Using the camera <{camera.__class__.__name__}>')
        self.cam = self.device
        self.cam.servertitle = f'Server on port {port}'
        self.cam.wait_for_client = self.wait_for_client
        
        added_functions = {'acquireSeries': self.cam.acquire_series,
                          'acquireSingle': self.cam.acquire_single,
                          'saveDescription': self.cam.save_description,
                          'set_roi': self.cam.set_roi,
                          'set_save_stack': self.cam.set_save_stack,
                          'get_cameras': self.cam.get_cameras,
                          'get_camera': self.cam.get_camera,
                          'set_camera': self.cam.set_camera,
                          'get_settings': self.cam.get_settings,
                          'get_setting_type': self.cam.get_setting_type,
                          'get_setting': self.cam.get_setting,
                          'set_setting': self.cam.set_setting,
                          }

        self.functions = {**self.functions, **added_functions}

        self.responders.extend(
                ['get_cameras', 'get_camera', 'get_settings',
                 'get_setting_type', 'get_setting']
                )

        

def test_camera():
    cam = Camera()
    images = cam.acquireSeries(0.01, 1, 5, 'testing')
    
    for image in images:
        plt.imshow(image, cmap='gray')
        plt.show()



def main():
    
    parser = argparse.ArgumentParser(
            prog='GonioImsoft Camera Server',
            description='Controls a MicroManager camera')

    parser.add_argument('-p', '--port')
    parser.add_argument('-c', '--camera')
    parser.add_argument('-s', '--save-directory')

    args = parser.parse_args()


    if args.camera == 'mm':
        Camera = MMCamera
    elif args.camera == 'dummy':
        Camera = DummyCamera
    else:
        # Default
        if pymmcore:
            Camera = MMCamera
        else:
            Camera = DummyCamera

    camera = Camera()

    if args.save_directory:
        self.set_save_directory(args.save_directory)

    if args.port:
        args.port = int(args.port)

    cam_server = CameraServer(camera, args.port)
    cam_server.run()
            
        
if __name__ == "__main__":
    main()



'''Places for saving and loading files.

Constants
---------
USERDATA_DIR : string
    Path to the user data that contains small files
    such as settings.
'''

import os
import platform

CODE_ROOTDIR = os.path.dirname(os.path.realpath(__file__))
USER_HOMEDIR = os.path.expanduser('~')

if platform.system() == "Windows":
    USERDATA_DIR = os.path.join(USER_HOMEDIR, 'GonioImsoft')
else:
    USERDATA_DIR = os.path.join(USER_HOMEDIR, '.gonioimsoft')

if os.path.isdir(USERDATA_DIR):
    IS_USERDATA_INITIALIZED = True
else:
    IS_USERDATA_INITIALIZED = False

def initialize_userdata():
    '''Create all user data diretories used by GonioImsoft

    It is good to ask for the user's consent before running this
    function unless USERDATA_DIR exists already.
    '''
    dirs = [
            USERDATA_DIR,
            os.path.join(USERDATA_DIR, 'macros'),
            os.path.join(USERDATA_DIR, 'biosyst_stimuli'),
            os.path.join(USERDATA_DIR, 'presets'),
            os.path.join(USERDATA_DIR, 'camera_states'),
            ]
    for adir in dirs:
        os.makedirs(adir, exist_ok=True)
